/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_AshcraftLiu.h - modification of Primitive moves refinement algorithm by Ashcraft and Liu
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/


#include "morsy_AshcraftLiu.h"
#include "math.h"

//-----------------------------------------------------------------------
//Archraft Liu Primitive moves
//-----------------------------------------------------------------------

void AshcraftLiuProcedureOuterWeighted(GraphStructure* Graph, 
                                       PartitionStructure* BestPartition, 
                                       PartitionStructure* TempPartition,
                                       NestedDissectionParameters* params, 
                                       int* isLooked, int* AllMovedVertexes, 
                                       int maxOuterLoops)
{
  int i;
  int j;
  int moveSep = 0;
  int smallPart;
  int bigPart;
  int nLooked;
  int nMoved = 0;
  int nAllMoved = 0;
  int neighbor;
  int nAllMovedNew = 0;
  int iter = 0;
  int inneriter;
  int nVert = Graph->nVertexes;
  int* GraphXadj = Graph->Xadj;
  int* GraphAdjncy = Graph->Adjncy;
  int gainCut = (1 - Graph->maxAdjWeight) / params->gainCutLimit; 
  FLOAT_TYPE oldBestEval;
  int xadjStart, xadjEnd;
  int maxInnerLoops = nVert;
  int bestIter;
  int bestGainCut;
  FLOAT_TYPE allowedDisbalance;
  int f = 0;
  TempPartition->Gains[0]->gainCut = gainCut;
  
  do
  {
    iter++;
    allowedDisbalance = params->disbalanceStop * BestPartition->disbalance;
    CopyPartition(BestPartition, TempPartition, nVert);
    oldBestEval = EvaluatePartitionWeighted(BestPartition, params);
    ReinitializeGainsStructure(TempPartition->Gains[0], nVert);
    memset(isLooked, 0, sizeof(int)*nVert);

    for(i = 0; i < TempPartition->nSep; i++)
    {
      TempPartition->SeparatorListPos[TempPartition->SeparatorList[i]] = i;
    }

    if (TempPartition->firstPartWeight > TempPartition->secondPartWeight)
    {  
      smallPart = pSecondPart; 
      bigPart = pFirstPart; 
    }
    else 
    {
      smallPart = pFirstPart; 
      bigPart = pSecondPart; 
    }

    for (i = 0; i < TempPartition->nSep; i++)
    {
      moveSep = TempPartition->SeparatorList[i];
      //s->V1
      InsertGain(TempPartition->Gains[0], moveSep, 
        GetGainWeighted(TempPartition, moveSep, bigPart, Graph));
    }

    nAllMoved = 0;
    inneriter = 0;
    nLooked = 0;
    bestIter = 0;
    bestGainCut = TempPartition->Gains[0]->maxGain;

    while ((TempPartition->Gains[0]->maxGain > gainCut) &&
           (TempPartition->disbalance < allowedDisbalance))
    {
      inneriter++;
      moveSep = FindBestPrimitiveMove(TempPartition->Gains[0]);

      nMoved = UpdatePartitionWeighted(moveSep, TempPartition, Graph, smallPart, 
        AllMovedVertexes + nAllMoved);
      AllMovedVertexes[nAllMoved + nMoved] = moveSep;
      isLooked[moveSep] = 1;
      nLooked++;

      UpdateAdjToPartWeights(moveSep, TempPartition, Graph, smallPart, 
        AllMovedVertexes + nAllMoved, nMoved);
      
      if ((TempPartition->nFirstPart > 0) && (TempPartition->nSecondPart > 0))
      {
        if (EvaluatePartitionWeighted(TempPartition, params) < 
            EvaluatePartitionWeighted(BestPartition, params))
        {
          CopyPartitionForList(TempPartition, BestPartition, AllMovedVertexes, 
            nAllMoved + nMoved + 1, Graph);
          nAllMovedNew = 0;
          bestIter = inneriter;
        }
        else 
        {
          nAllMovedNew = nAllMoved + nMoved + 1;
        }

        for (i = 0; i < nMoved; i++)
        {
          InsertGain(TempPartition->Gains[0], AllMovedVertexes[nAllMoved + i], 
            GetGainWeighted(TempPartition, AllMovedVertexes[nAllMoved + i], 
            bigPart, Graph));
        }

        for (i = 0; i < nMoved; i++)
        {
          xadjStart = GraphXadj[AllMovedVertexes[nAllMoved + i]];
          xadjEnd   = GraphXadj[AllMovedVertexes[nAllMoved + i] + 1];
          for (j = xadjStart; j < xadjEnd; j++)
          {
            neighbor = GraphAdjncy[j];
            if (isLooked[neighbor] + TempPartition->Partition[neighbor] == 0)
            {
              UpdateGain(TempPartition->Gains[0], neighbor, 
                GetGainWeighted(TempPartition, neighbor, bigPart, Graph));
            }
          }
        } // nMoved
        nAllMoved = nAllMovedNew;

        if (inneriter == bestIter)
        {
          bestGainCut = TempPartition->Gains[0]->maxGain;
        }
      } // if 
      else break;
    }//inner loop
  }
  while ((EvaluatePartitionWeighted(BestPartition, params) < oldBestEval) 
      && (iter < maxOuterLoops)); 
}

void AshcraftLiuProcedureOuterWeightedUnlim(GraphStructure* Graph, 
                                       PartitionStructure* BestPartition, 
                                       PartitionStructure* TempPartition,
                                       NestedDissectionParameters* params, 
                                       int* isLooked, int* AllMovedVertexes, 
                                       int maxOuterLoops)
{
  int i;
  int j;
  int moveSep = 0;
  int smallPart;
  int bigPart;
  int nLooked;
  int nMoved = 0;
  int nAllMoved = 0;
  int neighbor;
  int nAllMovedNew = 0;
  int iter = 0;
  int inneriter;
  int nVert = Graph->nVertexes;
  int* GraphXadj = Graph->Xadj;
  int* GraphAdjncy = Graph->Adjncy;
  int gainCut = (1 - Graph->maxAdjWeight) / 3; 
  FLOAT_TYPE oldBestEval;
  int xadjStart, xadjEnd;
  int maxInnerLoops = nVert;
  int* partW [2];

  TempPartition->Gains[0]->gainCut = gainCut;

  do
  {
    iter++;

    CopyPartition(BestPartition, TempPartition, nVert);
    oldBestEval = EvaluatePartitionWeighted(BestPartition, params);
    ReinitializeGainsStructure(TempPartition->Gains[0], nVert);
    memset(isLooked, 0, sizeof(int)*nVert);

    for(i = 0; i < TempPartition->nSep; i++)
    {
      TempPartition->SeparatorListPos[TempPartition->SeparatorList[i]] = i;
    }

    if (TempPartition->firstPartWeight > TempPartition->secondPartWeight)
    {  
      smallPart = pSecondPart; partW[0] = &TempPartition->secondPartWeight;
      bigPart = pFirstPart; partW[1] = &TempPartition->firstPartWeight;
    }
    else 
    {
      smallPart = pFirstPart; partW[0] = &TempPartition->firstPartWeight;
      bigPart = pSecondPart; partW[1] = &TempPartition->secondPartWeight;
    }

    for (i = 0; i < TempPartition->nSep; i++)
    {
      moveSep = TempPartition->SeparatorList[i];
      //s->V1
      InsertGain(TempPartition->Gains[0], moveSep, 
        GetGainWeighted(TempPartition, moveSep, bigPart, Graph));
    }

    nAllMoved = 0;
    inneriter = 0;
    nLooked = 0;

    while ((TempPartition->Gains[0]->maxGain > gainCut)  
        && (nLooked < maxInnerLoops))
    {
      inneriter++;
      moveSep = FindBestPrimitiveMove(TempPartition->Gains[0]);

      nMoved = UpdatePartitionWeighted(moveSep, TempPartition, Graph, smallPart, 
        AllMovedVertexes + nAllMoved);
      AllMovedVertexes[nAllMoved + nMoved] = moveSep;
      isLooked[moveSep] = 1;
      nLooked++;

      UpdateAdjToPartWeights(moveSep, TempPartition, Graph, smallPart, 
        AllMovedVertexes + nAllMoved, nMoved);
      
      if ((TempPartition->nFirstPart > 0) && (TempPartition->nSecondPart > 0))
      {
        if (EvaluatePartitionWeighted(TempPartition, params) < 
            EvaluatePartitionWeighted(BestPartition, params))
        {
          CopyPartitionForList(TempPartition, BestPartition, AllMovedVertexes, 
            nAllMoved + nMoved + 1, Graph);
          nAllMovedNew = 0;
        }
        else 
        {
          nAllMovedNew = nAllMoved + nMoved + 1;
        }

        for (i = 0; i < nMoved; i++)
        {
          InsertGain(TempPartition->Gains[0], AllMovedVertexes[nAllMoved + i], 
            GetGainWeighted(TempPartition, AllMovedVertexes[nAllMoved + i], 
            bigPart, Graph));
        }

        for (i = 0; i < nMoved; i++)
        {
          xadjStart = GraphXadj[AllMovedVertexes[nAllMoved + i]];
          xadjEnd   = GraphXadj[AllMovedVertexes[nAllMoved + i] + 1];
          for (j = xadjStart; j < xadjEnd; j++)
          {
            neighbor = GraphAdjncy[j];
            if (isLooked[neighbor] + TempPartition->Partition[neighbor] == 0)
            {
              UpdateGain(TempPartition->Gains[0], neighbor, 
                GetGainWeighted(TempPartition, neighbor, bigPart, Graph));
            }
          }
        } // nMoved
        nAllMoved = nAllMovedNew;
      } // if 
      else break;
    }//inner loop
  }
  while ((EvaluatePartitionWeighted(BestPartition, params) < oldBestEval) 
      && (iter < maxOuterLoops)); 
}


