/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_coarsening.c - coarsening rotines 
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/


#include "morsy_coarsening.h"
#include "pmorsy.h"
#include "omp.h"
#include "morsy_randomGenerator.h"

int OrderMatching(int* Match, int* Map, int* MatchRoots, int n)
{
  int i;
  int j = 0;
  for(i = 0; i < n; i++)
  {
    if (i <= Match[i])
    {
      Map[i] = j;
      Map[Match[i]] = j;
      MatchRoots[j] = i;
      j++;
    }
  }
  return j;
}

void GenerateAdjncyPermutation(GraphStructure* Graph, int* adjncyVisitOrder, 
                               MORSY_RNG_PTR stream)
{
  int i;
  int* randomVector = (int*)malloc(sizeof(int) * Graph->maxDegree);

  for(i = 0; i < Graph->nVertexes; i++)
  {
    GetRandomVector(stream, randomVector, Graph->Xadj[i + 1] - Graph->Xadj[i], 0, 
      Graph->Xadj[i + 1] - Graph->Xadj[i]);
    CreateRandomPermutationFromVector(adjncyVisitOrder + Graph->Xadj[i], 
      randomVector, Graph->Xadj[i + 1] - Graph->Xadj[i]);
  }

  free(randomVector);
}

int HeavyEdgeMatching(GraphStructure* Graph, int* adjncyVisitOrder, 
                      int* visitOrder, MORSY_RNG_PTR stream)
{
  int n = Graph->nVertexes;
  int nMatched = 0;
  int i;
  int j;
  int vertex;
  int neighbor;
  int max;
  int maxId;
  int xadjStart, xadjEnd;
  int* Match = Graph->Match;

  for(i = 0; i < n; i++)
  {
    vertex = visitOrder[i];
    if (Match[vertex] == vertex)
    {
      max = 0; maxId = -1;
      xadjStart = Graph->Xadj[vertex];
      xadjEnd   = Graph->Xadj[vertex + 1];
      for (j = xadjStart; j < xadjEnd; j++)
      {
        neighbor = Graph->Adjncy[j];
        if (Match[neighbor] == neighbor)
        {
          if (Graph->EdgeWeights[j] > max)
          {
            max = Graph->EdgeWeights[j];
            maxId = neighbor;
          }
        }
      }
      if (maxId > -1)
      {
        Match[maxId] = vertex;
        Match[vertex] = maxId;     
        nMatched++;
      }
    }
  }

  return nMatched;
}


int RandomMatching(GraphStructure* Graph, 
                   int* adjncyVisitOrder, int* visitOrder, 
                   MORSY_RNG_PTR stream)
{
  int n = Graph->nVertexes;
  int nMatched = 0;
  int i;
  int j;
  int vertex;
  int neighbor;
  int* Match = Graph->Match;

  for(i = 0; i < n; i++)
  {
    vertex = visitOrder[i];
    if (Match[vertex] == vertex)
    {
      for (j = Graph->Xadj[vertex]; j < Graph->Xadj[vertex + 1]; j++)
      {
        neighbor = Graph->Adjncy[Graph->Xadj[vertex] + adjncyVisitOrder[j]];
        if (Match[neighbor] == neighbor)
        {
          Match[neighbor] = vertex;
          Match[vertex] = neighbor;     
          nMatched++;
          break;
        }
      }
    }
  }

  return nMatched;
}

int HeavyCliqueMatching(GraphStructure* Graph, 
                    int* adjncyVisitOrder, int* visitOrder, 
                    MORSY_RNG_PTR stream)
{
  int n = Graph->nVertexes;
  int nMatched = 0;
  int i;
  int j;
  int vertex;
  int neighbor;
  FLOAT_TYPE max;
  int maxId;
  int xadjStart, xadjEnd;
  FLOAT_TYPE edgeDensity;
  int* Match = Graph->Match;

  for(i = 0; i < n; i++)
  {
    vertex = visitOrder[i];
    if (Match[vertex] == vertex)
    {
      max = 0.0; 
      maxId = -1;
      xadjStart = Graph->Xadj[vertex];
      xadjEnd   = Graph->Xadj[vertex + 1];
      for (j = xadjStart; j < xadjEnd; j++)
      {
        neighbor = Graph->Adjncy[j];
        if (Match[neighbor] == neighbor)
        {
          edgeDensity = 2.0 * (Graph->ContractedWeights[vertex] + 
             Graph->ContractedWeights[neighbor] + Graph->EdgeWeights[j]) / 
            (Graph->VertexWeights[vertex] + Graph->VertexWeights[neighbor]) / 
            (Graph->VertexWeights[vertex] + Graph->VertexWeights[neighbor] - 1.0);
          if (edgeDensity > max)
          {
            max = edgeDensity;
            maxId = neighbor;
          }
        }
      }
      if (maxId > -1)
      {
        Match[maxId] = vertex;
        Match[vertex] = maxId;     
        nMatched++;
      }
    }
  }

  return nMatched;

}

int ContractingGraph(GraphStructure* Graph, GraphStructure* coarserGraph,
                     int nMatched)
{
  int i;
  int j;
  int n = Graph->nVertexes;
  int nAdjncyEdges;
  int vertex; 
  int adjStart;
  int neighbor;
  int mapVertex;
  int mapNeighbor;
  int matchVertex;
  int* included = (int*)malloc(sizeof(int)*(Graph->nVertexes - nMatched));
  int* Match = Graph->Match;
  int* Map = Graph->Map;
  int k;
  int xadjStart, xadjEnd;

  MallocGraphStructure(coarserGraph, Graph->nVertexes - nMatched,
    Graph->nEdges - nMatched);

  memset(included, 0, sizeof(int)*coarserGraph->nVertexes);

  coarserGraph->isFirst = Graph->isFirst;
  coarserGraph->nVertexes = Graph->nVertexes - nMatched;
  coarserGraph->Xadj[0] = 0;
  coarserGraph->sumWeight = Graph->sumWeight;
  coarserGraph->maxVertexWeight = 0;
  coarserGraph->maxAdjWeight = 0;
  coarserGraph->maxDegree = 0;

  for (i = 0; i < Graph->nMatchRoots; i++)
  {
    vertex = Graph->MatchRoots[i];
    mapVertex = Map[vertex];
    matchVertex = Match[vertex];

    coarserGraph->VertexWeights[i] = Graph->VertexWeights[vertex];
    coarserGraph->ContractedWeights[i] = Graph->ContractedWeights[vertex];
    included[mapVertex] = 1;
    nAdjncyEdges = 0;
    adjStart = coarserGraph->Xadj[i];
    xadjStart = Graph->Xadj[vertex];
    xadjEnd = Graph->Xadj[vertex + 1];
    for(j = xadjStart; j < xadjEnd; j++)
    {
      neighbor = Graph->Adjncy[j];
      mapNeighbor = Map[neighbor];
      if (!included[mapNeighbor])
      {
        coarserGraph->Adjncy[adjStart + nAdjncyEdges] = mapNeighbor;
        nAdjncyEdges++;  
      }
      included[mapNeighbor] += Graph->EdgeWeights[j];
    }

    if (matchVertex!= vertex)
    {
      coarserGraph->VertexWeights[i] += Graph->VertexWeights[matchVertex];
      coarserGraph->ContractedWeights[i] += Graph->ContractedWeights[matchVertex];
      xadjStart = Graph->Xadj[matchVertex];
      xadjEnd   = Graph->Xadj[matchVertex + 1];
      for(j = xadjStart; j < xadjEnd; j++)
      {
        neighbor = Graph->Adjncy[j];
        mapNeighbor = Map[neighbor];
        if (!included[mapNeighbor])
        {
          coarserGraph->Adjncy[adjStart + nAdjncyEdges] = mapNeighbor;
          nAdjncyEdges++;
        }
        else if (neighbor == vertex)
        {
          coarserGraph->ContractedWeights[i] += Graph->EdgeWeights[j];
        }
        included[mapNeighbor] += Graph->EdgeWeights[j];
      }
    }

    coarserGraph->Xadj[i + 1] = adjStart + nAdjncyEdges;
    if (!nAdjncyEdges)
    {
      free(included);
      return 0;
    }

    xadjStart = coarserGraph->Xadj[i];
    xadjEnd   = coarserGraph->Xadj[i + 1];
    for (j = xadjStart; j < xadjEnd; j++)
    {
      coarserGraph->EdgeWeights[j] = included[coarserGraph->Adjncy[j]];
      included[coarserGraph->Adjncy[j]] = 0;
    }
    included[mapVertex] = 0;

    if (nAdjncyEdges > coarserGraph->maxDegree)
    {
      coarserGraph->maxDegree = nAdjncyEdges;
    }
    if (coarserGraph->VertexWeights[i] > coarserGraph->maxVertexWeight)
    {
      coarserGraph->maxVertexWeight = coarserGraph->VertexWeights[i];
    }
  } // for vertex

  for(i = 0; i < coarserGraph->nVertexes; i++)
  {
    xadjStart = coarserGraph->Xadj[i];
    xadjEnd   = coarserGraph->Xadj[i + 1];
    for (j = xadjStart; j < xadjEnd; j++)
    {
      coarserGraph->AdjncyWeights[i] += 
        coarserGraph->VertexWeights[coarserGraph->Adjncy[j]];
    }
    if (coarserGraph->AdjncyWeights[i] > coarserGraph->maxAdjWeight)
    {
      coarserGraph->maxAdjWeight = coarserGraph->AdjncyWeights[i];
    }
  }

  coarserGraph->nEdges = coarserGraph->Xadj[coarserGraph->nVertexes];

  free(included);
  return 1;
}


int Coarsening(GraphStructure* Graph, GraphStructure** CoarsenedQueue,
                NestedDissectionParameters* params)
{
  int n = Graph->nVertexes;
  int* visitOrder = (int*)malloc(sizeof(int)*n);
  int* randomVector = (int*)malloc(sizeof(int)*n);
  int* adjncyVisitOrder = (int*)malloc(sizeof(int)*Graph->nEdges);
  int i;
  int j;
  int nMatched;
  int isConnected;
  int nCoarseningSteps;
  int threadID = omp_get_thread_num();

  nCoarseningSteps = params->nMaxCoarseningSteps;
  CoarsenedQueue[0] = Graph;
  for(i = 0; i < params->nMaxCoarseningSteps; i++)
  {
#ifdef MORSY_TIMERS
    StartTimer(&(params->timer[MORSY_TIME_MATCHING]));
    if (Graph->isFirst)
    {
      StartTimer(&(params->timer[MORSY_TIME_FIRST_TASK_MATCHING]));
    }
#endif
    for(j = 0; j < CoarsenedQueue[i]->nVertexes; j++)
    {
      CoarsenedQueue[i]->Match[j] = j;
    }

    GetRandomVector(params->randomStream[threadID], randomVector, CoarsenedQueue[i]->nVertexes, 0,
      CoarsenedQueue[i]->nVertexes);

    CreateRandomPermutationFromVector(visitOrder, randomVector, 
      CoarsenedQueue[i]->nVertexes);

    if (params->matchingType == 0)
    {
      GenerateAdjncyPermutation(CoarsenedQueue[i], adjncyVisitOrder, params->randomStream[threadID]);
    }

    nMatched = params->Matching(CoarsenedQueue[i],  
      adjncyVisitOrder, visitOrder, params->randomStream[threadID]);

    CoarsenedQueue[i]->nMatchRoots = OrderMatching(CoarsenedQueue[i]->Match, CoarsenedQueue[i]->Map, 
      CoarsenedQueue[i]->MatchRoots, CoarsenedQueue[i]->nVertexes);

#ifdef MORSY_TIMERS
    FinishTimer(&(params->timer[MORSY_TIME_MATCHING]));
    if (Graph->isFirst)
    {
      FinishTimer(&(params->timer[MORSY_TIME_FIRST_TASK_MATCHING]));
    }
    StartTimer(&(params->timer[MORSY_TIME_CONTRACTING]));
#endif
    
    isConnected = ContractingGraph(CoarsenedQueue[i], CoarsenedQueue[i + 1], 
      nMatched);

#ifdef MORSY_TIMERS
    FinishTimer(&(params->timer[MORSY_TIME_CONTRACTING]));
#endif

    if (!isConnected)
    {
      nCoarseningSteps = i;
      FreeGraphStructure(CoarsenedQueue[i + 1]);
      break;
    }

    if ((CoarsenedQueue[i + 1]->nEdges <= CoarsenedQueue[i + 1]->nVertexes - 2) ||
      (CoarsenedQueue[i + 1]->nVertexes < params->nMinVertexes * 2))
    {
      nCoarseningSteps = i + 1;
      break;
    }
  }

  free(adjncyVisitOrder);
  free(visitOrder);
  free(randomVector);

  return nCoarseningSteps;
}
