/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_compression.h - matrix compression 
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/


#include "morsy_compression.h"

//-----------------------------------------------------------------------
//GRAPH COMPRESSION ROUTINES
//-----------------------------------------------------------------------

int AdjncyHashValue(int vertex, int* GraphAdjncy, int* GraphXadj)
{
  int hash = vertex;
  int i;
  for (i = GraphXadj[vertex]; i < GraphXadj[vertex + 1]; i++)
    hash += GraphAdjncy[i];
  return hash;
}

int AdjncyComparison(int vertex, int* GraphAdjncy, int* GraphXadj, 
                     int* AdjncyMask)
{
  int isEqual = 1;
  int i;
  for (i = GraphXadj[vertex]; i < GraphXadj[vertex + 1]; i++)
  {
    if (!AdjncyMask[GraphAdjncy[i]]) 
    {
      isEqual = 0;
      break;
    }
  }
  return isEqual;
}

void MallocCompressionStructure(CompressionStructure* Compressed, int n)
{
  Compressed->nOriginal = n;
  Compressed->merged = (int*)malloc(sizeof(int)*n);
  Compressed->node = (int*)malloc(sizeof(int)*n);
  Compressed->nCompressedWithIt = (int*)malloc(sizeof(int)*n);
  Compressed->nodeNames = (int*)malloc(sizeof(int)*n);
  memset(Compressed->nCompressedWithIt, 0, sizeof(int)*n);
}

int GraphCompression(int* GraphXadj, int* GraphAdjncy, int* VertexMarks, 
                     CompressionStructure* Compressed, int n, 
                     int* nExtraEdges)
{
  int nCompr = 0;
  int i;
  int j;
  int neigb;
  int rootHash;
  int neighbHash;
  int isEqual;
  int* AdjncyMask;
  int* hashValues;

  int xadjStart;
  int xadjEnd;

  AdjncyMask = (int*)malloc(sizeof(int)*n);
  memset(AdjncyMask, 0, sizeof(int)*n);
  hashValues = (int*)malloc(sizeof(int)*n);

  for(i = 0; i < n; i++)
  {
	  hashValues[i] = AdjncyHashValue(i, GraphAdjncy, GraphXadj);
  }

  *nExtraEdges = 0;
  for (i = 0; i < n; i++)
  {
    if (!VertexMarks[i])
    {
      rootHash = hashValues[i];
      xadjStart = GraphXadj[i];
      xadjEnd   = GraphXadj[i + 1];
      
      for (j = xadjStart; j < xadjEnd; j++)
        AdjncyMask[GraphAdjncy[j]] = 1;
      AdjncyMask[i] = 1;

      for (j = xadjStart; j < xadjEnd; j++)
      {
        neigb = GraphAdjncy[j];

        if (!VertexMarks[neigb])
        {  
          neighbHash = hashValues[neigb];
          if ((rootHash == neighbHash) && 
              (xadjEnd - xadjStart == GraphXadj[neigb + 1] - GraphXadj[neigb]))
          {
            isEqual = AdjncyComparison(neigb, GraphAdjncy, GraphXadj, AdjncyMask);
            if (isEqual) 
            {
              Compressed->merged[nCompr] = neigb;
              Compressed->node[nCompr] = i;
              VertexMarks[neigb] = 1;
              nCompr++;
              *nExtraEdges += GraphXadj[neigb + 1] - GraphXadj[neigb];
              Compressed->nCompressedWithIt[i]++;
            }
          }
        }
      }

      for (j = xadjStart; j < xadjEnd; j++)
        AdjncyMask[GraphAdjncy[j]] = 0;
      AdjncyMask[i] = 0;
    }
  }
  Compressed->nCompressed = nCompr;
  (*nExtraEdges) -= nCompr;

  free(hashValues);
  free(AdjncyMask);
  return nCompr;
}


void CompressAdjncyLists(int* VertexMarks, int* GraphXadj, int* GraphAdjncy, 
                         int** NewGraphAdjncy, int** NewGraphXadj, 
                         CompressionStructure* Compressed,
                         int nExtraEdges, int n)
{
  int i;
  int j;
  int pos = 0;
  int nodePos = 0;
  int xadjStart;
  int xadjEnd;

  (*NewGraphXadj) = (int*)malloc(sizeof(int)*
                                 (n + 1 - Compressed->nCompressed));
  (*NewGraphAdjncy) = (int*)malloc(sizeof(int)*(GraphXadj[n] - nExtraEdges));
  
  for(i = 0; i < n; i++)
  {
    if (!VertexMarks[i])
    {
      Compressed->nodeNames[i] = nodePos++;
    }
    else 
    {
      Compressed->nodeNames[i] = -1;
    }
  }
  
  nodePos = 0;
  for(i = 0; i < n; i++)
  {
    if (!VertexMarks[i])
    {
      (*NewGraphXadj)[nodePos++] = pos;
      xadjStart = GraphXadj[i];
      xadjEnd   = GraphXadj[i + 1];
      for(j = xadjStart; j < xadjEnd; j++)
      {
        if (!VertexMarks[GraphAdjncy[j]])
        {
          (*NewGraphAdjncy)[pos++] = Compressed->nodeNames[GraphAdjncy[j]];
        }
      }
    }
  }
  (*NewGraphXadj)[nodePos] = pos;
}

void FreeCompressionStructure(CompressionStructure* Compressed)
{
  free(Compressed->merged);
  free(Compressed->nCompressedWithIt);
  free(Compressed->node);
  free(Compressed->nodeNames);
}

int GraphCompressionProcedure(int* GraphAdjncy, int* GraphXadj,
                              int** NewGraphAdjncy, int** NewGraphXadj,
                              CompressionStructure* Compressed, int n) 
{
  int nCompressed;
  int nExtraEdges;
  int* VertexMarks;

  VertexMarks = (int*)malloc(sizeof(int)*n);
  memset(VertexMarks, 0, sizeof(int)*n);
  MallocCompressionStructure(Compressed, n);

  nCompressed = GraphCompression(GraphXadj, GraphAdjncy, VertexMarks, 
    Compressed, n, &nExtraEdges);
  
  if (nCompressed > 0)
  {
    CompressAdjncyLists(VertexMarks, GraphXadj, GraphAdjncy, NewGraphAdjncy, 
      NewGraphXadj, Compressed, nExtraEdges, n);
  }
  else
  {
    *NewGraphXadj = (int*)malloc(sizeof(int) * (n + 1));
    memcpy(*NewGraphXadj, GraphXadj,  sizeof(int) * (n + 1));
    *NewGraphAdjncy = (int*)malloc( sizeof(int) * GraphXadj[n]);
    memcpy(*NewGraphAdjncy, GraphAdjncy, sizeof(int) * GraphXadj[n]);
  }

  free(VertexMarks);

  return nCompressed;
}