/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_gainsOperations.c - operations with gains structure
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/


#include "morsy_gainsOperations.h"

void InitializeGainsStructure(GainsStructure* Gains, int n)
{
  int i;
  for(i = 0; i < Gains->nBuckets; i++)
  {  
    Gains->Buckets[i] = -1;
  }
  Gains->maxGain = Gains->minPossGain;
  Gains->oldMaxGain = Gains->minPossGain;
  Gains->nNodes = 0;
  for (i = 0; i < n; i++)
  {  
    Gains->Nodes[i].next = -1;
    Gains->Nodes[i].prev = -1;
    Gains->gainValue[i] = Gains->minPossGain - 1;
  }
}

 void ReinitializeGainsStructure(GainsStructure* Gains, int n)
{
  int i;
  for(i = 0; i < Gains->nBuckets; i++)
  {  
    Gains->Buckets[i] = -1;
  }
  for (i = 0; i < n; i++)
  {  
    Gains->Nodes[i].next = -1;
    Gains->Nodes[i].prev = -1;
    Gains->gainValue[i] = Gains->minPossGain - 1;
  }
  Gains->maxGain = Gains->minPossGain;
  Gains->oldMaxGain = Gains->minPossGain;
  Gains->nNodes = 0;
}

void MallocGainsStructure(GainsStructure* Gains, int minGain, int maxGain, 
                          int n)
{
  int i;

  Gains->nBuckets = maxGain - minGain + 1;
  Gains->minPossGain = minGain;

  Gains->Nodes = (DoublyLinkedList*)malloc(sizeof(DoublyLinkedList)*n);
  Gains->Buckets = (int*)malloc(sizeof(int) * Gains->nBuckets);
  Gains->gainValue = (int*)malloc(sizeof(int) * n);
}

void FreeGainsStructure(GainsStructure* Gains)
{
  free(Gains->Nodes);
  free(Gains->Buckets);
  free(Gains->gainValue);
}


 void InsertGain(GainsStructure* Gains, int node, int gain)
{
  DoublyLinkedList* newElem;
  int gainLoc = -Gains->minPossGain + gain;

  if (gain < Gains->gainCut)
    return;

  newElem = &Gains->Nodes[node];
  Gains->gainValue[node] = gain;

  newElem->next = Gains->Buckets[gainLoc];
  newElem->prev = -1;

  if (newElem->next != -1)
    Gains->Nodes[newElem->next].prev = node;

  Gains->Buckets[gainLoc] = node;

  if (Gains->maxGain < gain)
    Gains->maxGain = gain;
  Gains->nNodes++;
}

 void DeleteGain(GainsStructure* Gains, int node)
{
  DoublyLinkedList* newElem;
  int index = -Gains->minPossGain;
  int oldgain;  
  newElem = &Gains->Nodes[node];
  oldgain = Gains->gainValue[node];

  if (oldgain < Gains->minPossGain)
    return;

  Gains->gainValue[node] = Gains->minPossGain - 1;

  if (newElem->prev != -1)
    Gains->Nodes[newElem->prev].next = newElem->next;
  else 
    Gains->Buckets[index + oldgain] = newElem->next;

  if (newElem->next != -1)
    Gains->Nodes[newElem->next].prev = newElem->prev;

  Gains->nNodes--;

  if (oldgain == Gains->maxGain)
  {
    while (Gains->Buckets[Gains->maxGain + index] == -1)
      Gains->maxGain--;
  }
}

 void UpdateGain(GainsStructure* Gains, int node, int newGain)
{
  DoublyLinkedList* newElem = &(Gains->Nodes[node]);
  int index = -Gains->minPossGain;
  int oldGain = Gains->gainValue[node];

  if (oldGain == newGain)
    return;

  //deleting
  if (oldGain >= Gains->minPossGain)
  {
    if (newElem->prev != -1)
      Gains->Nodes[newElem->prev].next = newElem->next;
    else 
        Gains->Buckets[index + oldGain] = newElem->next;

    if (newElem->next != -1)
      Gains->Nodes[newElem->next].prev = newElem->prev;

    Gains->nNodes--;
  }

  if (newGain < Gains->gainCut)
  {
    if (oldGain == Gains->maxGain)
    {
      while (Gains->Buckets[Gains->maxGain + index] == -1)
        Gains->maxGain--;
    }
    return;
  }

  //inserting
  Gains->nNodes++;
  Gains->gainValue[node] = newGain;
  newElem->next = Gains->Buckets[index + newGain];
  newElem->prev = -1;

  if (newElem->next != -1)
    Gains->Nodes[newElem->next].prev = node;
  Gains->Buckets[index + newGain] = node;

  if (Gains->maxGain < newGain)
    Gains->maxGain = newGain;
  else
    if (oldGain == Gains->maxGain)
    {
      while (Gains->Buckets[Gains->maxGain + index] == -1)
        Gains->maxGain--;
    }
}

int FindBestPrimitiveMove(GainsStructure* Gains)
{
  DoublyLinkedList* node;
  int vertex = -1;
  int index = -Gains->minPossGain;

  if (Gains->nNodes == 0)
    return vertex;

  Gains->oldMaxGain = Gains->maxGain;

  vertex = Gains->Buckets[Gains->maxGain + index];
  node = &Gains->Nodes[vertex];
  Gains->gainValue[vertex] = Gains->minPossGain - 1;
  Gains->nNodes--;

  Gains->Buckets[Gains->maxGain + index] = node->next;
  if (node->next != -1) 
    Gains->Nodes[node->next].prev = -1;
  else 
  {
    if (Gains->nNodes > 0)
    {
      while (Gains->Buckets[Gains->maxGain + index] == -1)
        Gains->maxGain--;
    }
    else
    {
      Gains->maxGain = -index;
    }
  }    

  return vertex;
}

int ViewBestPrimitiveMove(GainsStructure* Gains)
{
  if (Gains->nNodes == 0)
    return -1;
  else 
    return Gains->Buckets[Gains->maxGain - Gains->minPossGain];
}
