/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_initPartitioning.c - routines for initial graph partitioning
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/


#include "morsy_initPartitioning.h"
#include "morsy_routines.h"
#include "morsy_partitionOperations.h"
#include "morsy_gainsOperations.h"
#include "morsy_randomGenerator.h"
#include "math.h"

void GreedyGraphGrowingPartitionFromVertex(PartitionStructure* Partition,  
                                GraphStructure* Graph, int root);
void KerniganLinPartitionRefinement(PartitionStructure* Partition,  
                                GraphStructure* Graph);

void BFSPartition(PartitionStructure* Partition,  
                        GraphStructure* Graph,  MORSY_RNG_PTR stream)
{
  int* LevelStructure;
  int* LevelIndex;
  int n = Graph->nVertexes;
  int partLevel;
  int levelStart;
  int levelEnd;
  int vertex;
  int nLevels;
  int i;
  int j;

  LevelStructure = (int*)malloc(sizeof(int)*n);
  LevelIndex     = (int*)malloc(sizeof(int)*(n + 1));

  partLevel = FindSeparatorWeighted(Graph, LevelStructure, LevelIndex, 
    Partition->SeparatorList, &(Partition->nSep));

  if (Partition->nSep == Graph->nVertexes)
  {
    for (i = 0; i < Graph->nVertexes; i++)
    {
      Partition->Partition[i] = pSeparator;
    }
    Partition->firstPartWeight = 0;
    Partition->secondPartWeight = 0;
    Partition->nFirstPart = 0;
    Partition->nSecondPart = 0;
    Partition->sepWeight = Graph->sumWeight;
    Partition->sumWeight = Graph->sumWeight;
  }
  else
  {
    Partition->firstPartWeight = 0;
    Partition->secondPartWeight = 0;
    Partition->sepWeight = 0;

    levelStart = LevelIndex[partLevel];
    levelEnd   = LevelIndex[partLevel + 1];

    for (i = 0; i < levelStart; i++)
    {  
      vertex = LevelStructure[i];
      Partition->Partition[vertex] = pFirstPart;
      Partition->firstPartWeight += Graph->VertexWeights[vertex];
    }

    if (Partition->nSep < levelEnd - levelStart)
    {
      j = 0;
      for (i = levelStart; i < levelEnd; i++)
      {
        vertex = LevelStructure[i];
        if (j < Partition->nSep)
        {
          if (LevelStructure[i] == Partition->SeparatorList[j])
          {
            Partition->Partition[vertex] = pSeparator;
            Partition->sepWeight += Graph->VertexWeights[vertex];
            j++;
          }
          else 
          {
            Partition->Partition[vertex] = pFirstPart;
            Partition->firstPartWeight += Graph->VertexWeights[vertex];
          }
        }
        else
        {
          Partition->Partition[vertex] = pFirstPart;
          Partition->firstPartWeight += Graph->VertexWeights[vertex];
        }
      }
    }
    else
    {
      for (i = levelStart; i < levelEnd; i++)
      {  
        vertex = LevelStructure[i];
        Partition->Partition[vertex] = pSeparator;
        Partition->sepWeight += Graph->VertexWeights[vertex];
      }
    }

    for (i = levelEnd; i < Graph->nVertexes; i++)
    {
      vertex = LevelStructure[i];
      Partition->Partition[vertex] = pSecondPart;
      Partition->secondPartWeight += Graph->VertexWeights[vertex];
    }

    Partition->nSecondPart = Graph->nVertexes - levelEnd;
    Partition->nVert = Graph->nVertexes;
    Partition->nFirstPart = levelEnd - Partition->nSep; 
    Partition->sumWeight = Partition->firstPartWeight + Partition->secondPartWeight 
      + Partition->sepWeight;

    if (Partition->firstPartWeight > Partition->secondPartWeight)
    {
      Partition->disbalance = 
        (FLOAT_TYPE)Partition->firstPartWeight / 
        (FLOAT_TYPE)Partition->secondPartWeight;
    }
    else
    {
      Partition->disbalance = 
        (FLOAT_TYPE)Partition->secondPartWeight / 
        (FLOAT_TYPE)Partition->firstPartWeight;
    }
  }

  free(LevelIndex);
  free(LevelStructure);
}


void GraphGrowingPartition(PartitionStructure* Partition,  
                           GraphStructure* Graph,  MORSY_RNG_PTR stream) // #1
{
  int* VertexMarks;
  int* LevelStructure; 
  int* LevelIndex;
  PartitionStructure* TestPartition;
  int* roots;
  int i;
  int n = Graph->nVertexes;
  int nRuns = 10;
  int nLevels;
  int min = n + 1;
  int minID;

  VertexMarks = (int*)malloc(sizeof(int) * n);
  LevelStructure = (int*)malloc(sizeof(int) * n);
  LevelIndex     = (int*)malloc(sizeof(int) * (n + 1));
  TestPartition = (PartitionStructure*)malloc(sizeof(PartitionStructure) * nRuns);
  roots = (int*)malloc(sizeof(int) * nRuns);

  GetRandomVector(stream, roots, nRuns, 0, n);

  for (i = 0; i < nRuns; i++)
  {
    InitializePartitionMemoryLight(&TestPartition[i], n, 0);
  }

  for (i = 0; i < Graph->nVertexes; i++)
    VertexMarks[i] = 1;

  for (i = 0; i < nRuns; i++)
  {
    nLevels = RootedLevelStructure(roots[i], Graph->Xadj, Graph->Adjncy, VertexMarks,
      LevelStructure, LevelIndex);
    PartitionFromLevelStructure(LevelStructure, LevelIndex, nLevels, 
      &TestPartition[i], Graph);
    VertexSeparatorFromEdgeSeparator(&TestPartition[i], Graph, 0);

    if (TestPartition[i].nSep < min)
    {
      min = TestPartition[i].nSep;
      minID = i;
    }
  }

  CopyPartitionWithSepList(&TestPartition[minID], Partition);
  InitializePartitionParameters(Partition, Graph);

  for (i = 0; i < nRuns; i++)
  {
    FreePartitionMemoryLight(&TestPartition[i]);
  }
  free(VertexMarks);
  free(LevelStructure);
  free(LevelIndex);
  free(roots);
  free(TestPartition);
}


void InitPartition(PartitionStructure* Partition)
{
  int i;
  int n = Partition->nVert;

  for (i = 0; i < n; i++)
  {
    Partition->Partition[i] = pSecondPart;
  }
}

void GreedyGraphGrowingPartition(PartitionStructure* Partition,  
                                GraphStructure* Graph,  MORSY_RNG_PTR stream) // #2
{
  PartitionStructure* TestPartition;
  int* roots;
  int i;
  int j;
  int n = Graph->nVertexes;
  int nRuns = 10;
  int minID;
  int minGain;
  int maxGain;
  int minSep = n + 1;

  TestPartition = (PartitionStructure*)malloc(sizeof(PartitionStructure) * nRuns);
  roots = (int*)malloc(sizeof(int) * nRuns);

  GetRandomVector(stream, roots, nRuns, 0, n);

  GetMinMaxEdgeGain(Graph, &minGain, &maxGain);
  for (i = 0; i < nRuns; i++)
  {
    InitializePartitionMemory(&TestPartition[i], n, -maxGain, maxGain, 1);
  }

  for (i = 0; i < nRuns; i++)
  {
    InitPartition(&TestPartition[i]);
  }

  GetAdjToPartWeightsForEdgePartition(&TestPartition[0], Graph);
  for (i = 1; i < nRuns; i++)
  {
    CopyAdjToPartWeights(&TestPartition[0], &TestPartition[i]);
  }

  for (i = 0; i < nRuns; i++)
  {
    GreedyGraphGrowingPartitionFromVertex(&TestPartition[i], Graph, roots[i]);    
    VertexSeparatorFromEdgeSeparator(&TestPartition[i], Graph, 1);
    
    if (TestPartition[i].nSep < minSep)
    {
      minSep = TestPartition[i].nSep;
      minID = i;
    }
  }

  CopyPartitionWithSepList(&TestPartition[minID], Partition);
  InitializePartitionParameters(Partition, Graph);

  for (i = 0; i < nRuns; i++)
  {
    FreePartitionMemory(&TestPartition[i]);
  }
  free(roots);
  free(TestPartition);
}

void KerniganLinPartition(PartitionStructure* Partition,  
                                GraphStructure* Graph,  MORSY_RNG_PTR stream) // #3
{
  int* VertexMarks;
  int* LevelStructure; 
  int* LevelIndex;
  PartitionStructure* TestPartition;
  int* roots;
  int i;
  int n = Graph->nVertexes;
  int nRuns = 5;
  int nLevels;
  int min = n + 1;
  int minID;
  int minGain;
  int maxGain;

  VertexMarks = (int*)malloc(sizeof(int) * n);
  LevelStructure = (int*)malloc(sizeof(int) * n);
  LevelIndex     = (int*)malloc(sizeof(int) * (n + 1));
  TestPartition = (PartitionStructure*)malloc(sizeof(PartitionStructure) * nRuns);
  roots = (int*)malloc(sizeof(int) * nRuns);

  GetRandomVector(stream, roots, nRuns, 0, n);

  GetMinMaxEdgeGain(Graph, &minGain, &maxGain);
  for (i = 0; i < nRuns; i++)
  {
    InitializePartitionMemory(&TestPartition[i], n, -maxGain, maxGain, 2);
  }

  for (i = 0; i < Graph->nVertexes; i++)
    VertexMarks[i] = 1;

  for (i = 0; i < nRuns; i++)
  {
    nLevels = RootedLevelStructure(roots[i], Graph->Xadj, Graph->Adjncy, VertexMarks,
      LevelStructure, LevelIndex);
    PartitionFromLevelStructure(LevelStructure, LevelIndex, nLevels, 
      &TestPartition[i], Graph);
    
    GetAdjToPartWeightsForEdgePartition(&TestPartition[i], Graph);
    InitializePartitionParameters(&TestPartition[i], Graph);
    KerniganLinPartitionRefinement(&TestPartition[i], Graph);

    VertexSeparatorFromEdgeSeparator(&TestPartition[i], Graph, 0);

    if (TestPartition[i].nSep < min)
    {
      min = TestPartition[i].nSep;
      minID = i;
    }
  }

  CopyPartitionWithSepList(&TestPartition[minID], Partition);
  InitializePartitionParameters(Partition, Graph);

  for (i = 0; i < nRuns; i++)
  {
    FreePartitionMemory(&TestPartition[i]);
  }
  free(VertexMarks);
  free(LevelStructure);
  free(LevelIndex);
  free(roots);
  free(TestPartition);
}


void GreedyGraphGrowingPartitionFromVertex(PartitionStructure* Partition,  
                                GraphStructure* Graph, int root)
{
  int i;
  int movedVertex;
  int neighbor;
  int xadjStart;
  int xadjEnd;
  int n = Graph->nVertexes;
  int weight = 0;
  int barrier = Graph->sumWeight / 2;

  ReinitializeGainsStructure(Partition->Gains[0], n);
  Partition->Gains[0]->gainCut = Partition->Gains[0]->minPossGain; 
  Partition->Gains[0]->oldMaxGain = -Partition->AdjToPartWeights[1][root];
  weight = Graph->VertexWeights[root];
  UpdateEdgePartitionAndGains(root, Partition, Graph);
  
  while (weight < barrier)
  {
    movedVertex = FindBestPrimitiveMove(Partition->Gains[0]);
    UpdateEdgePartitionAndGains(movedVertex, Partition, Graph);
    weight += Graph->VertexWeights[movedVertex];
  }

}

void KerniganLinPartitionRefinement(PartitionStructure* Partition,  
                                    GraphStructure* Graph)
{
  int i;
  int movedVertex;
  int neighbor;
  int xadjStart;
  int xadjEnd;
  int n = Graph->nVertexes;
  int* isLooked;
  int* bestPartition;
  int* allMoved;
  int nAllMoved = 0;
  int nLooked = 0;
  int startEdgeCut;
  int startDisbalance;
  int part;
  int bigPart;
  FLOAT_TYPE allowedDisbalance;
  int lastSuccsessfulIteration = 0;
  int iteration = 0;

  isLooked = (int*)malloc(sizeof(int) * n);
  bestPartition = (int*)malloc(sizeof(int) * n);
  allMoved = (int*)malloc(sizeof(int) * n);

  ReinitializeGainsStructure(Partition->Gains[0], n);
  ReinitializeGainsStructure(Partition->Gains[1], n);
  Partition->Gains[0]->gainCut = Partition->Gains[0]->minPossGain;
  Partition->Gains[1]->gainCut = Partition->Gains[1]->minPossGain;

  startDisbalance = Partition->disbalance;
  startEdgeCut = Partition->edgeCut;
  allowedDisbalance = Partition->disbalance * 1.2;

  memcpy(bestPartition, Partition->Partition, sizeof(int)*n);
  memset(isLooked, 0, sizeof(int)*n);

  for (i = 0; i < n; i++)
  {
    part = Partition->Partition[i] - 1;
    InsertGain(Partition->Gains[part], i, GetEdgeGain(Partition, i));
  }

  bigPart = (Partition->firstPartWeight > Partition->secondPartWeight) ? (pFirstPart - 1) : (pSecondPart - 1);

  while ((nLooked < n) && (Partition->Gains[bigPart]->maxGain > Partition->Gains[0]->gainCut) && 
         (Partition->disbalance < allowedDisbalance) && (iteration - lastSuccsessfulIteration < n / 2))
  {
    movedVertex = ViewBestPrimitiveMove(Partition->Gains[bigPart]);
    isLooked[movedVertex] = 1;
    UpdateEdgePartitionAndGainsForAll(movedVertex, Partition, Graph, isLooked);
    nLooked++;

    allMoved[nAllMoved] = movedVertex;
    nAllMoved++;

    if (Partition->edgeCut < startEdgeCut) 
    {
      lastSuccsessfulIteration = iteration;
      startEdgeCut = Partition->edgeCut;
      UpdateArrayForVertexesList(bestPartition, Partition->Partition, allMoved, nAllMoved);
      nAllMoved = 0;
    }
    bigPart = (Partition->firstPartWeight > Partition->secondPartWeight) ? (pFirstPart - 1) : (pSecondPart - 1);

    iteration++;
  }

  if (iteration > lastSuccsessfulIteration)
  {
    UpdateArrayForVertexesList(Partition->Partition, bestPartition, allMoved, nAllMoved);
  }

  free(isLooked);
  free(bestPartition);
  free(allMoved);
}