/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_multilevelND.c - multilevel nested dissection scheme
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/


#include "morsy_multilevelND.h"
#include "omp.h"

void NDStep(TaskStructure* task, AshcraftLiuWeighted AshcraftLiuProcedure,
            NestedDissectionParameters* params, int isRecursive)
{
  GraphStructure* graph = task->graph;
  GraphStructure** graphQueue;
  GraphStructure* coarserGraph;

  PartitionStructure* coarserPartition;
  PartitionStructure* finerPartition;
  int i;
  int n = graph->nVertexes;
  int startTaskID;
  int nCoarseningSteps;
  int k;
  int sepLevel;
  int myid;

  if (graph->nVertexes > params->nMinVertexes)
  {
#ifdef MORSY_TIMERS
    if (graph->isFirst)
    {
      StartTimer(&(params->timer[MORSY_TIME_FIRST_TASK]));
    }
#endif

    graphQueue = 
      (GraphStructure**)malloc(sizeof(GraphStructure*) * 
      (params->nMaxCoarseningSteps + 1));
    coarserPartition = (PartitionStructure*)malloc(sizeof(PartitionStructure));
    finerPartition = (PartitionStructure*)malloc(sizeof(PartitionStructure));

    for(i = 1; i < params->nMaxCoarseningSteps + 1; i++)
    {
      graphQueue[i] = (GraphStructure*)malloc(sizeof(GraphStructure));
    }

#ifdef MORSY_TIMERS
    if (graph->isFirst)
    {
      StartTimer(&(params->timer[MORSY_TIME_FIRST_TASK_COARSENING]));
    }
    StartTimer(&(params->timer[MORSY_TIME_COARSENING]));
#endif
    nCoarseningSteps = Coarsening(graph, graphQueue, params);
#ifdef MORSY_TIMERS
    FinishTimer(&(params->timer[MORSY_TIME_COARSENING]));
    if (graph->isFirst)
    {
      FinishTimer(&(params->timer[MORSY_TIME_FIRST_TASK_COARSENING]));
    }
#endif
    coarserGraph = graphQueue[nCoarseningSteps];
#ifdef MORSY_TIMERS
    StartTimer(&(params->timer[MORSY_TIME_INIT_PARTITIONING]));
    if (graph->isFirst)
    {
      StartTimer(&(params->timer[MORSY_TIME_FIRST_TASK_INIT_PARTITIONING]));
    }
#endif

    InitializePartitionMemoryLight(finerPartition, n, 1);
    InitializePartitionMemoryLight(coarserPartition, coarserGraph->nVertexes, 1);

    params->Partitioning(coarserPartition, coarserGraph, params->randomStream[omp_get_thread_num()]);

#ifdef MORSY_TIMERS
    FinishTimer(&(params->timer[MORSY_TIME_INIT_PARTITIONING]));
    if (graph->isFirst)
    {
      FinishTimer(&(params->timer[MORSY_TIME_FIRST_TASK_INIT_PARTITIONING]));
    }
    StartTimer(&(params->timer[MORSY_TIME_UNCOARSENING]));
    if (graph->isFirst)
    {
      StartTimer(&(params->timer[MORSY_TIME_FIRST_TASK_UNCOARSENING]));
    }
#endif

    Refinement(graphQueue, coarserPartition, finerPartition, 
      AshcraftLiuProcedure, params, nCoarseningSteps);

#ifdef MORSY_TIMERS
    FinishTimer(&(params->timer[MORSY_TIME_UNCOARSENING]));
    if (graph->isFirst)
    {
      FinishTimer(&(params->timer[MORSY_TIME_FIRST_TASK_UNCOARSENING]));
    }
#endif
    for (i = 1; i < nCoarseningSteps + 1; i++)
    {
      FreeGraphStructure(graphQueue[i]);
    }
    for(i = 1; i < params->nMaxCoarseningSteps + 1; i++)
    {
      free(graphQueue[i]);
    }

    task->nSep = finerPartition->nSep;
    task->Separators = (int*)malloc(sizeof(int) * task->nSep);
    for (i = 0; i < finerPartition->nSep; i++)
    {
      task->Separators[i] = 
        graph->GlobalMap[finerPartition->SeparatorList[i]];
    }
#ifdef MORSY_TIMERS
    StartTimer(&(params->timer[MORSY_TIME_SPLITTING]));
#endif
    task->nSubgraphs = FindSubgraphs(graph, finerPartition, &task->subgraphs);
#ifdef MORSY_TIMERS
    FinishTimer(&(params->timer[MORSY_TIME_SPLITTING]));

    if (graph->isFirst)
    {
      FinishTimer(&(params->timer[MORSY_TIME_FIRST_TASK]));
    }
#endif

    task->subTasks = (TaskStructure**)malloc(sizeof(TaskStructure*) * task->nSubgraphs);

    for (i = 0; i < task->nSubgraphs; i++)
    {
      task->subTasks[i] = (TaskStructure*)malloc(sizeof(TaskStructure));
      task->subTasks[i]->graph = task->subgraphs[i];
      task->subTasks[i]->level = task->level + 1;
      task->subTasks[i]->part = task->subgraphs[i]->part;
    }

    FreePartitionMemoryLight(coarserPartition);
    FreePartitionMemoryLight(finerPartition);
    free(coarserPartition);
    free(finerPartition);
    free(graphQueue);

    for (i = 0; i < task->nSubgraphs; i++)
    {
      if (task->subTasks[i]->graph->nVertexes > params->barrier)
      {
#pragma omp task   
        NDStep(task->subTasks[i], AshcraftLiuProcedure, params, 1); 
      }
      else
        NDStep(task->subTasks[i], AshcraftLiuProcedure, params, 1); 
    }
  }
  else
  { // SMALL GRAPH
#ifdef MORSY_TIMERS
    StartTimer(&(params->timer[MORSY_TIME_LEAFS]));
    if (graph->isFirst)
    {
      StartTimer(&(params->timer[MORSY_TIME_FIRST_TASK]));
    }
#endif    
    
    task->Separators = (int*)malloc(sizeof(int) * graph->nVertexes);
    BasicNestedDissectionProcedure(graph->Xadj, graph->Adjncy, graph->nVertexes,
      task->Separators);

    for (i = 0; i < graph->nVertexes; i++)
    {
      task->Separators[i] = graph->GlobalMap[task->Separators[i]];
    }
    task->nSep = graph->nVertexes;
    task->nSubgraphs = 0;
#ifdef MORSY_TIMERS
    FinishTimer(&(params->timer[MORSY_TIME_LEAFS]));
    if (graph->isFirst)
    {
      FinishTimer(&(params->timer[MORSY_TIME_FIRST_TASK]));
    }
#endif  
  }

  FreeGraphStructure(graph);
  free(graph);
}

void CollectSeparators(TaskStructure* task, int* Separators, int* k)
{
  int i;

  memcpy(Separators + (*k), task->Separators, sizeof(int) * task->nSep);
  (*k) += task->nSep;
  for (i = 0; i < task->nSubgraphs; i++)
  {
    CollectSeparators(task->subTasks[i], Separators, k);
  }
}

void FreeTasksMemory(TaskStructure* task)
{
  int i;

  free(task->Separators);
  for (i = 0; i < task->nSubgraphs; i++)
  {
    FreeTasksMemory(task->subTasks[i]);
    free(task->subTasks[i]);
  }
  if (task->nSubgraphs > 0)
  {
    free(task->subgraphs);
    free(task->subTasks);
  }
}

void MultilevelNestedDissection(GraphStructure* Graph, int* Separators,
                                AshcraftLiuWeighted AshcraftLiuProcedure, 
                                NestedDissectionParameters* params)
{
  TaskStructure* task = (TaskStructure*)malloc(sizeof(TaskStructure));
  int* locSeparators;
  int i, taskStart, taskEnd, sumTaskW;
  int nLoc;

  task->graph = Graph;
  task->level = 0;

#pragma omp parallel
  {
#pragma omp single
    {
      NDStep(task, AshcraftLiuProcedure, params, 1);
    }  
  }
  nLoc = 0;
  CollectSeparators(task, Separators, &nLoc);
  FreeTasksMemory(task);
  free(task);
}