/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_partitionOperations.c - bacis operations with partition structure
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/


#include "morsy_partitionOperations.h"
#include "math.h"
#include "morsy_routines.h"

void InitializePartitionMemory(PartitionStructure* Partition, int n, 
                               int minGain, int maxGain, int nGains)
{
  int i;

  Partition->nGains = nGains;
  Partition->nVert = n;

  InitializePartitionMemoryLight(Partition, n, 1);
  Partition->SeparatorListPos = (int*)malloc(sizeof(int) * n);

  for (i = 0; i < nGains; i++)
  {
    Partition->Gains[i] = (GainsStructure*)malloc(sizeof(GainsStructure));
    MallocGainsStructure(Partition->Gains[i], minGain, maxGain, n);
  }
}

void InitializePartitionMemoryLight(PartitionStructure* Partition, int n, int isAdjToPartNeed)
{
  Partition->nVert = n;
  Partition->Partition = 
    (int*)malloc(n*sizeof(int));
  Partition->SeparatorList = (int*)malloc(n * sizeof(int));

  if (isAdjToPartNeed)
  {
    Partition->AdjToPartWeights[0] = (int*)malloc(sizeof(int) * n);
    Partition->AdjToPartWeights[1] = (int*)malloc(sizeof(int) * n);
  }
  else
  {
    Partition->AdjToPartWeights[0] = NULL;
    Partition->AdjToPartWeights[1] = NULL;
  }
  
}

void FreePartitionMemory(PartitionStructure* Partition)
{
  int i;

  for (i = 0; i < Partition->nGains; i++)
  {
    FreeGainsStructure(Partition->Gains[i]);
    free(Partition->Gains[i]);
  }
  free(Partition->SeparatorListPos);
  FreePartitionMemoryLight(Partition);
}

void FreePartitionMemoryLight(PartitionStructure* Partition)
{
  free(Partition->Partition);
  free(Partition->SeparatorList);

  if (Partition->AdjToPartWeights[0])
  {
    free(Partition->AdjToPartWeights[0]);
  }
  if (Partition->AdjToPartWeights[1])
  {
    free(Partition->AdjToPartWeights[1]);
  }
}

void CopyPartitionParameters(PartitionStructure* PartitionFrom, 
                             PartitionStructure* PartitionTo)
{
  PartitionTo->nFirstPart = PartitionFrom->nFirstPart;
  PartitionTo->nSecondPart = PartitionFrom->nSecondPart;
  PartitionTo->firstPartWeight = PartitionFrom->firstPartWeight;
  PartitionTo->secondPartWeight = PartitionFrom->secondPartWeight;
  PartitionTo->sepWeight = PartitionFrom->sepWeight;
  PartitionTo->nSep = PartitionFrom->nSep;
  PartitionTo->nVert = PartitionFrom->nVert;
  PartitionTo->sumWeight = PartitionFrom->sumWeight;
  PartitionTo->disbalance = PartitionFrom->disbalance;
}

void CopyPartitionForList(PartitionStructure* PartitionFrom, 
                          PartitionStructure* PartitionTo, int* Vertexes,
                          int nVert, GraphStructure* Graph)
{
  int i;
  int j;
  int vertex;
  int neighbor;
  int n = PartitionFrom->nVert;

  memcpy(PartitionTo->SeparatorList, PartitionFrom->SeparatorList, 
    sizeof(int)*PartitionFrom->nSep);

  for (i = 0; i < nVert; i++)
  {    
    vertex = Vertexes[i];
    PartitionTo->Partition[vertex] = PartitionFrom->Partition[vertex];
    PartitionTo->AdjToPartWeights[0][vertex] = PartitionFrom->AdjToPartWeights[0][vertex];
    PartitionTo->AdjToPartWeights[1][vertex] = PartitionFrom->AdjToPartWeights[1][vertex];
    for (j = Graph->Xadj[vertex]; j < Graph->Xadj[vertex + 1]; j++)
    {
      neighbor = Graph->Adjncy[j];
      PartitionTo->AdjToPartWeights[0][neighbor] = PartitionFrom->AdjToPartWeights[0][neighbor];
      PartitionTo->AdjToPartWeights[1][neighbor] = PartitionFrom->AdjToPartWeights[1][neighbor];
    }
  }
  CopyPartitionParameters(PartitionFrom, PartitionTo);
}


void CopyPartition(PartitionStructure* PartitionFrom, 
                   PartitionStructure* PartitionTo, int nVert)
{
  memcpy(PartitionTo->SeparatorList, PartitionFrom->SeparatorList, 
    sizeof(int) * PartitionFrom->nSep);
  memcpy(PartitionTo->Partition, PartitionFrom->Partition, sizeof(int)*nVert);
  memcpy(PartitionTo->AdjToPartWeights[0], PartitionFrom->AdjToPartWeights[0], 
    sizeof(int) * nVert);
  memcpy(PartitionTo->AdjToPartWeights[1], PartitionFrom->AdjToPartWeights[1], 
    sizeof(int) * nVert);

  CopyPartitionParameters(PartitionFrom, PartitionTo);
}

void CopyPartitionWithSepList(PartitionStructure* PartitionFrom, 
                   PartitionStructure* PartitionTo)
{
  memcpy(PartitionTo->SeparatorList, PartitionFrom->SeparatorList, 
    sizeof(int) * PartitionFrom->nSep);
  memcpy(PartitionTo->Partition, PartitionFrom->Partition, sizeof(int)*PartitionFrom->nVert);
}

FLOAT_TYPE EvaluatePartitionWeighted(PartitionStructure* Partition,
                                     NestedDissectionParameters* params)
{
    return (FLOAT_TYPE)Partition->sepWeight * 
    (1.0 + params->disbalanceCoeff * Partition->disbalance);
}

int GetGainWeighted(PartitionStructure* Partition, int moveSep, 
                    int bigPart, GraphStructure* Graph)
{
  return Graph->VertexWeights[moveSep] - 
    Partition->AdjToPartWeights[bigPart - 1][moveSep];
}

int GetEdgeGain(PartitionStructure* Partition, int vertex)
{
  int myPart = Partition->Partition[vertex] - 1;

  return Partition->AdjToPartWeights[!myPart][vertex] - 
         Partition->AdjToPartWeights[ myPart][vertex];
}

int UpdatePartitionWeighted(int moveSep, PartitionStructure* Partition,  
                            GraphStructure* Graph, int part, 
                            int* MovedVertexes)

{
  int i;
  int sepPos;
  int neighbor;
  int nMoved = 0;
  int otherPart = part % 2 + 1;
  int movedVertexesWeight = 0;
  int xadjStart, xadjEnd;

  sepPos = Partition->SeparatorListPos[moveSep];

  Partition->Partition[moveSep] = part;
  Partition->SeparatorList[sepPos] = 
    Partition->SeparatorList[Partition->nSep - 1];
  Partition->SeparatorListPos[Partition->SeparatorList[sepPos]] = sepPos;
  Partition->sepWeight -= Graph->VertexWeights[moveSep];
  Partition->nSep--;

  xadjStart = Graph->Xadj[moveSep];
  xadjEnd   = Graph->Xadj[moveSep + 1];
  for (i = xadjStart; i < xadjEnd; i++)
  {
    neighbor = Graph->Adjncy[i];
    if (Partition->Partition[neighbor] == otherPart)
    {
      Partition->SeparatorList[Partition->nSep] = neighbor;
      Partition->SeparatorListPos[neighbor] = Partition->nSep;
      Partition->Partition[neighbor] = pSeparator;
      Partition->nSep++;
      movedVertexesWeight += Graph->VertexWeights[neighbor];
      MovedVertexes[nMoved] = neighbor;
      nMoved++;
    }
  }

  Partition->sepWeight += movedVertexesWeight;

  if (part == pFirstPart)
  {
    Partition->nFirstPart++;
    Partition->nSecondPart -= nMoved;
    Partition->firstPartWeight += Graph->VertexWeights[moveSep];
    Partition->secondPartWeight -= movedVertexesWeight;
  }
  else
  {
    Partition->nFirstPart -= nMoved;
    Partition->nSecondPart++;
    Partition->firstPartWeight -= movedVertexesWeight;
    Partition->secondPartWeight += Graph->VertexWeights[moveSep];
  }

  if (Partition->firstPartWeight > Partition->secondPartWeight)
  {
    Partition->disbalance = (FLOAT_TYPE)Partition->firstPartWeight / 
      (FLOAT_TYPE)Partition->secondPartWeight;
  }
  else
  {
    Partition->disbalance = (FLOAT_TYPE)Partition->secondPartWeight / 
      (FLOAT_TYPE)Partition->firstPartWeight;
  }

  return nMoved;
}

void UpdateAdjToPartWeights(int moveSep, PartitionStructure* Partition,  
                            GraphStructure* Graph, int partToMove, 
                            int* MovedVertexes, int nMoved)
{
  int i;
  int j;
  int vertex;
  int n = Graph->nVertexes;
  int* part1 = Partition->AdjToPartWeights[partToMove - 1];
  int* part2 = Partition->AdjToPartWeights[partToMove % 2];
  int xadjStart = Graph->Xadj[moveSep];
  int xadjEnd   = Graph->Xadj[moveSep + 1];
  int dweight = Graph->VertexWeights[moveSep];

  for(i = xadjStart; i < xadjEnd; i++)
  {
    part1[Graph->Adjncy[i]] += dweight;
  }

  for (i = 0; i < nMoved; i++)
  {
    vertex = MovedVertexes[i];
    xadjStart = Graph->Xadj[vertex];
    xadjEnd   = Graph->Xadj[vertex + 1];
    dweight = Graph->VertexWeights[vertex];
    for (j = xadjStart; j < xadjEnd; j++)
    {
      part2[Graph->Adjncy[j]] -= dweight;
    }
  }
}

void GetAdjToPartWeightForVertex(int vertex, GraphStructure* Graph, PartitionStructure* Partition)
{
  int i;
  int neighbor;
  int xadjStart = Graph->Xadj[vertex];
  int xadjEnd   = Graph->Xadj[vertex + 1];

  for (i = xadjStart; i < xadjEnd; i++)
  {
    neighbor = Graph->Adjncy[i];
    if (Partition->Partition[neighbor] > 0)
    {
      Partition->AdjToPartWeights[Partition->Partition[neighbor] - 1][vertex] += 
        Graph->VertexWeights[neighbor];
    }
  }
}

void GetAdjToPartWeights(PartitionStructure* Partition, GraphStructure* Graph)
{
  int i;
  int j;
  int n = Graph->nVertexes;
  int* partition = Partition->Partition;

  memset(Partition->AdjToPartWeights[0], 0, sizeof(int) * n);
  memset(Partition->AdjToPartWeights[1], 0, sizeof(int) * n);
  for (i = 0; i < n; i++)
  {
    GetAdjToPartWeightForVertex(i, Graph, Partition);
  }
}

void GetAdjToPartWeightsForEdgePartition(PartitionStructure* Partition, GraphStructure* Graph)
{
  int i;
  int j;
  int xadjStart;
  int xadjEnd;
  int n = Graph->nVertexes;
  int neighborPart;
  int myPart;

  memset(Partition->AdjToPartWeights[0], 0, sizeof(int) * n);
  memset(Partition->AdjToPartWeights[1], 0, sizeof(int) * n);
  Partition->edgeCut = 0;  
  
  for (i = 0; i < n; i++)
  {
    myPart = Partition->Partition[i] - 1;
    xadjStart = Graph->Xadj[i];
    xadjEnd   = Graph->Xadj[i + 1];
    for(j = xadjStart; j < xadjEnd; j++)
    {
      neighborPart = Partition->Partition[Graph->Adjncy[j]] - 1;
      Partition->AdjToPartWeights[neighborPart][i] += Graph->EdgeWeights[j];

      if ((neighborPart != myPart) && (i < Graph->Adjncy[j]))
      {
        Partition->edgeCut += Graph->EdgeWeights[j];
      }
    }
  }
}


void CopyAdjToPartWeights(PartitionStructure* PartitionFrom, PartitionStructure* PartitionTo)
{
  memcpy(PartitionTo->AdjToPartWeights[0], PartitionFrom->AdjToPartWeights[0], sizeof(int)*PartitionFrom->nVert);
  memcpy(PartitionTo->AdjToPartWeights[1], PartitionFrom->AdjToPartWeights[1], sizeof(int)*PartitionFrom->nVert);
}


void GetMinMaxEdgeGain(GraphStructure* Graph, int* minGain, int* maxGain)
{
  int n = Graph->nVertexes;
  int i;
  int j;
  int xadjStart;
  int xadjEnd;
  int* AdjEdgeWeight;

  AdjEdgeWeight = (int*)malloc(sizeof(int) * n);
  memset(AdjEdgeWeight, 0, sizeof(int) * n);

  for (i = 0; i < n; i++)
  {
    xadjStart = Graph->Xadj[i];
    xadjEnd   = Graph->Xadj[i + 1];
    for(j = xadjStart; j < xadjEnd; j++)
    {
      AdjEdgeWeight[i] += Graph->EdgeWeights[j];
    }
  }
  GetIntArrayMaxAndMin(AdjEdgeWeight, n, maxGain, minGain);

  free(AdjEdgeWeight);
}



void ProjectAdjToPartWeights(GraphStructure* coarserGraph, GraphStructure* finerGraph, 
                             PartitionStructure* coarserPartition, 
                             PartitionStructure* finerPartition)
{
  int i;
  int j;
  int vertexPart;
  int otherPart;
  int match;
  int oldVertex;

  memset(finerPartition->AdjToPartWeights[0], 0, sizeof(int)*finerGraph->nVertexes);
  memset(finerPartition->AdjToPartWeights[1], 0, sizeof(int)*finerGraph->nVertexes);

  for (i = 0; i < finerGraph->nVertexes; i++)
  {
    match = finerGraph->Match[i];
    vertexPart = finerPartition->Partition[i];
    otherPart = vertexPart % 2 + 1;
    oldVertex = finerGraph->Map[i];

    if (vertexPart > 0) // not a separator
    {
      if (coarserPartition->AdjToPartWeights[vertexPart - 1][oldVertex] == 
        coarserGraph->AdjncyWeights[oldVertex]) // inner vertex
      {
        finerPartition->AdjToPartWeights[vertexPart - 1][i] = finerGraph->AdjncyWeights[i];
      }
      else if (coarserPartition->AdjToPartWeights[vertexPart - 1][oldVertex] == 0) 
      {
        if  (i != match)
        {
          finerPartition->AdjToPartWeights[vertexPart - 1][i] = finerGraph->VertexWeights[match];
        }
      }
      else
      {
        GetAdjToPartWeightForVertex(i, finerGraph, finerPartition);
      }
    }
    else // is separator
    {
      if (coarserPartition->AdjToPartWeights[0][oldVertex] + 
        coarserPartition->AdjToPartWeights[1][oldVertex] > 0) // boundary separator
      {
        GetAdjToPartWeightForVertex(i, finerGraph, finerPartition);
      }
    }
  }
}

void UpdateEdgePartitionAndGains(int movedVertex, PartitionStructure* Partition,  
                                 GraphStructure* Graph)
{
  int i;
  int neighbor;
  int newPart = !(Partition->Partition[movedVertex] - 1);
  int xadjStart = Graph->Xadj[movedVertex];
  int xadjEnd   = Graph->Xadj[movedVertex + 1];

  Partition->Partition[movedVertex] = (int)(newPart + 1);
  Partition->edgeCut -= Partition->Gains[0]->oldMaxGain;

  for (i = xadjStart; i < xadjEnd; i++)
  {
    neighbor = Graph->Adjncy[i];
    Partition->AdjToPartWeights[ newPart][neighbor] += Graph->EdgeWeights[i];
    Partition->AdjToPartWeights[!newPart][neighbor] -= Graph->EdgeWeights[i];
    
    if (Partition->Partition[neighbor] != Partition->Partition[movedVertex])
    {
      UpdateGain(Partition->Gains[0], neighbor, 
        GetEdgeGain(Partition, neighbor));
    }
    else
    {
      DeleteGain(Partition->Gains[0], neighbor);
    }
  }
}

void UpdateEdgePartitionAndGainsForBoundary(int movedVertex, 
                                            PartitionStructure* Partition,  
                                            GraphStructure* Graph, 
                                            int* isLooked)
{
  int i;
  int neighbor;
  int newGain;
  int neighborPart;
  int newPart = !(Partition->Partition[movedVertex] - 1);
  int oldPart = !newPart;
  int xadjStart = Graph->Xadj[movedVertex];
  int xadjEnd   = Graph->Xadj[movedVertex + 1];

  Partition->Partition[movedVertex] = (int)(newPart + 1);
  Partition->edgeCut -= Partition->Gains[oldPart]->maxGain;
  
  if (newPart == 0)
  {
    Partition->firstPartWeight += Graph->VertexWeights[movedVertex];
    Partition->secondPartWeight -= Graph->VertexWeights[movedVertex];
    Partition->nFirstPart += 1;
    Partition->nSecondPart -= 1;
  }
  else
  {
    Partition->firstPartWeight -= Graph->VertexWeights[movedVertex];
    Partition->secondPartWeight += Graph->VertexWeights[movedVertex];
    Partition->nFirstPart += 1;
    Partition->nSecondPart -= 1;
  }
  Partition->disbalance = (Partition->firstPartWeight > Partition->secondPartWeight) ? 
    ((FLOAT_TYPE)Partition->firstPartWeight / (FLOAT_TYPE)Partition->secondPartWeight) : 
  ((FLOAT_TYPE)Partition->secondPartWeight / (FLOAT_TYPE)Partition->firstPartWeight);

  DeleteGain(Partition->Gains[oldPart], movedVertex);

  for (i = xadjStart; i < xadjEnd; i++)
  {
    neighbor = Graph->Adjncy[i];
    neighborPart = Partition->Partition[neighbor] - 1;

    Partition->AdjToPartWeights[newPart][neighbor] += Graph->EdgeWeights[i];
    Partition->AdjToPartWeights[oldPart][neighbor] -= Graph->EdgeWeights[i];
    
    if ((!isLooked[neighbor]) && (Partition->AdjToPartWeights[!neighborPart][neighbor]) > 0)
    {
      newGain = GetEdgeGain(Partition, neighbor);
      UpdateGain(Partition->Gains[neighborPart], neighbor, newGain);
    }
    else
    {
      DeleteGain(Partition->Gains[neighborPart], neighbor);
    }
  }
}

void UpdateEdgePartitionAndGainsForAll(int movedVertex, 
                                            PartitionStructure* Partition,  
                                            GraphStructure* Graph, 
                                            int* isLooked)
{
  int i;
  int neighbor;
  int newGain;
  int neighborPart;
  int newPart = !(Partition->Partition[movedVertex] - 1);
  int oldPart = !newPart;
  int xadjStart = Graph->Xadj[movedVertex];
  int xadjEnd   = Graph->Xadj[movedVertex + 1];

  Partition->Partition[movedVertex] = (int)(newPart + 1);
  Partition->edgeCut -= Partition->Gains[oldPart]->maxGain;
  
  if (newPart == 0)
  {
    Partition->firstPartWeight += Graph->VertexWeights[movedVertex];
    Partition->secondPartWeight -= Graph->VertexWeights[movedVertex];
    Partition->nFirstPart += 1;
    Partition->nSecondPart -= 1;
  }
  else
  {
    Partition->firstPartWeight -= Graph->VertexWeights[movedVertex];
    Partition->secondPartWeight += Graph->VertexWeights[movedVertex];
    Partition->nFirstPart += 1;
    Partition->nSecondPart -= 1;
  }
  Partition->disbalance = (Partition->firstPartWeight > Partition->secondPartWeight) ? 
    ((FLOAT_TYPE)Partition->firstPartWeight / (FLOAT_TYPE)Partition->secondPartWeight) : 
  ((FLOAT_TYPE)Partition->secondPartWeight / (FLOAT_TYPE)Partition->firstPartWeight);

  DeleteGain(Partition->Gains[oldPart], movedVertex);

  for (i = xadjStart; i < xadjEnd; i++)
  {
    neighbor = Graph->Adjncy[i];
    neighborPart = Partition->Partition[neighbor] - 1;

    Partition->AdjToPartWeights[newPart][neighbor] += Graph->EdgeWeights[i];
    Partition->AdjToPartWeights[oldPart][neighbor] -= Graph->EdgeWeights[i];
    
    if (!isLooked[neighbor])
    {
      newGain = GetEdgeGain(Partition, neighbor);
      UpdateGain(Partition->Gains[neighborPart], neighbor, newGain);
    }
  }
}