/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_randomGenerator.c - random numbers routines 
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/

#include "morsy_randomGenerator.h"

#ifdef MORSY_USE_MKL

#define MKL_GENERATOR       VSL_BRNG_MT2203
#define MKL_METHOD          VSL_RNG_METHOD_UNIFORM_STD
#define MKL_GAUSSIAN_METHOD VSL_RNG_METHOD_GAUSSIAN_BOXMULLER
#define MKL_SEED            7777777
#define MKL_GENERATOR_PORTION 1000




void GetRandomVector(MORSY_RNG_PTR stream, int* vector, int size, int low,
                     int high)
{
  int i;
  for(i = 0; i < size / MKL_GENERATOR_PORTION; i++)
  {
    viRngUniform(MKL_METHOD, stream, MKL_GENERATOR_PORTION, vector + i * MKL_GENERATOR_PORTION, low, high);
  }
  viRngUniform(MKL_METHOD, stream, size % MKL_GENERATOR_PORTION, vector + (size / MKL_GENERATOR_PORTION) * MKL_GENERATOR_PORTION,
    low, high);
}

void InitializeRandomGenerators(MORSY_RNG_PTR** streams, int nStreams)
{
  int i;
  *streams = (MORSY_RNG_PTR*)malloc(sizeof(MORSY_RNG_PTR)*nStreams);
  
  for (i = 0; i < nStreams; i++ )
  {
    vslNewStream(&(*streams)[i], MKL_GENERATOR + i, MKL_SEED);
  }
}

void FinalizeRandomGenerators(MORSY_RNG_PTR** streams, int nStreams)
{
  int i;
  for (i = 0; i < nStreams; i++)
  {
    vslDeleteStream(&(*streams)[i]);
  }
  free(*streams);
}


#else

#define RAND_SEED 500000

void GetRandomVector(MORSY_RNG_PTR stream, int* vector, int size, int low, int high)
{
  int i;
  for(i = 0; i < size; i++)
  {
    vector[i] = low + rand() % (high - low);
  }
}

void InitializeRandomGenerators(MORSY_RNG_PTR** streams, int nStreams)
{
  *streams = (MORSY_RNG_PTR*)malloc(sizeof(MORSY_RNG_PTR)*nStreams);
  srand(RAND_SEED);
}

void FinalizeRandomGenerators(MORSY_RNG_PTR** streams, int nStreams)
{
  free(*streams);
}



#endif

void CreateRandomPermutationFromVector(int* vector, int* randomVector, 
                                       int size)
{
  int i;
  int swappedValue;

  for(i = 0; i < size; i++)
  {
    vector[i] = i;
  }

  for(i = 0; i < size; i++)
  {
    swappedValue = vector[i];
    vector[i] = vector[randomVector[i]];
    vector[randomVector[i]] = swappedValue;
  }
}
