/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_refinement.c - refinement routines
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/

#include "morsy_refinement.h"

void ProjectPartition(GraphStructure* coarserGraph, GraphStructure* finerGraph, 
                      PartitionStructure* coarserPartition, 
                      PartitionStructure* finerPartition)
{
  int i;
  int sepPos = 0;
  int nNewVertexesInPart[3] = {0, 0, 0};

  CopyPartitionParameters(coarserPartition, finerPartition);

  for (i = 0; i < finerGraph->nVertexes; i++)
  {
    finerPartition->Partition[i] = 
      coarserPartition->Partition[finerGraph->Map[i]];
    if (finerGraph->Match[i] > i)
    {
       nNewVertexesInPart[finerPartition->Partition[i]]++;
    }
    if (finerPartition->Partition[i] == pSeparator)
    {
      finerPartition->SeparatorList[sepPos] = i;
      sepPos++;
    }
  }
  finerPartition->nFirstPart += nNewVertexesInPart[pFirstPart];
  finerPartition->nSecondPart += nNewVertexesInPart[pSecondPart];
  finerPartition->nSep += nNewVertexesInPart[pSeparator];
}

void Refinement(GraphStructure** CoarsenedQueue, 
              PartitionStructure* coarsestPartition, 
              PartitionStructure* finestPartition,
              AshcraftLiuWeighted AshcraftLiuProcedure,
              NestedDissectionParameters* params, int nCoarseningSteps)
{
  int i;
  int j;
  int k;
  PartitionStructure* tempPartition;
  PartitionStructure* coarserPartition;
  PartitionStructure* finerPartition;
  PartitionStructure* bufferPartition;

  int* isLooked;
  int* AllMovedVertexes;
  GraphStructure* Graph = CoarsenedQueue[0];
  GraphStructure* coarsestGraph = CoarsenedQueue[nCoarseningSteps];
  int nMin = coarsestGraph->nVertexes;
  int n = Graph->nVertexes;
  int minGain = 1 - coarsestGraph->maxAdjWeight;
  int maxGain = coarsestGraph->maxVertexWeight;
  int maxOuterLoops;

  tempPartition = (PartitionStructure*)malloc(sizeof(PartitionStructure));
  finerPartition = (PartitionStructure*)malloc(sizeof(PartitionStructure));
  coarserPartition = (PartitionStructure*)malloc(sizeof(PartitionStructure));

  InitializePartitionMemory(tempPartition, n, minGain, maxGain, 1); 
  InitializePartitionMemoryLight(finerPartition, n, 1);
  InitializePartitionMemoryLight(coarserPartition, n, 1);

         isLooked  = (int*)malloc(sizeof(int)*n);
  AllMovedVertexes = (int*)malloc(sizeof(int)*3*n);
  
  GetAdjToPartWeights(coarsestPartition, coarsestGraph);
   
  k = nCoarseningSteps % params->refinementStep;

  CopyPartition(coarsestPartition, coarserPartition, nMin);
#ifdef MORSY_TIMERS
  StartTimer(&(params->timer[MORSY_TIME_PROJECTION]));
#endif  
  for (i = nCoarseningSteps; i > nCoarseningSteps - k; i--)
  {
    ProjectPartition(CoarsenedQueue[i], CoarsenedQueue[i - 1], 
      coarserPartition, finerPartition);

    ProjectAdjToPartWeights(CoarsenedQueue[i], CoarsenedQueue[i - 1], 
      coarserPartition, finerPartition);

    bufferPartition = coarserPartition;
    coarserPartition = finerPartition;
    finerPartition = bufferPartition;
  }
#ifdef MORSY_TIMERS
  FinishTimer(&(params->timer[MORSY_TIME_PROJECTION]));
#endif
  for(; i > 0; i -= params->refinementStep)
  {
    if (params->refineLimit > 0)
    {
      maxOuterLoops = 1;
    }
    else
    {
      maxOuterLoops = CoarsenedQueue[i]->nVertexes;
    }
#ifdef MORSY_TIMERS
    StartTimer(&(params->timer[MORSY_TIME_REFINEMENT]));
    if (Graph->isFirst)
    {
      StartTimer(&(params->timer[MORSY_TIME_FIRST_TASK_REFINEMENT]));
    }
#endif
    AshcraftLiuProcedure(CoarsenedQueue[i], 
      coarserPartition, tempPartition, params, isLooked, AllMovedVertexes, 
      maxOuterLoops);
#ifdef MORSY_TIMERS
    FinishTimer(&(params->timer[MORSY_TIME_REFINEMENT]));
    if (Graph->isFirst)
    {
      FinishTimer(&(params->timer[MORSY_TIME_FIRST_TASK_REFINEMENT]));
    }
    StartTimer(&(params->timer[MORSY_TIME_PROJECTION]));
#endif

    for (j = 0; j < params->refinementStep; j++)
    {
      ProjectPartition(CoarsenedQueue[i - j], CoarsenedQueue[i - 1 - j], 
                       coarserPartition, finerPartition);

      ProjectAdjToPartWeights(CoarsenedQueue[i - j], CoarsenedQueue[i - 1 - j], 
                              coarserPartition, finerPartition);

      bufferPartition = coarserPartition;
      coarserPartition = finerPartition;
      finerPartition = bufferPartition;
    }
#ifdef MORSY_TIMERS
  FinishTimer(&(params->timer[MORSY_TIME_PROJECTION]));
#endif
  }

#ifdef MORSY_TIMERS
  StartTimer(&(params->timer[MORSY_TIME_REFINEMENT]));
  if (Graph->isFirst)
  {
    StartTimer(&(params->timer[MORSY_TIME_FIRST_TASK_REFINEMENT]));
  }
#endif
  AshcraftLiuProcedure(CoarsenedQueue[0], 
    coarserPartition, tempPartition, params, isLooked, AllMovedVertexes,
    CoarsenedQueue[0]->nVertexes);

#ifdef MORSY_TIMERS
  FinishTimer(&(params->timer[MORSY_TIME_REFINEMENT]));
  if (Graph->isFirst)
  {
    FinishTimer(&(params->timer[MORSY_TIME_FIRST_TASK_REFINEMENT]));
  }
#endif
  CopyPartition(coarserPartition, finestPartition, n);

  free(isLooked);
  free(AllMovedVertexes);

  FreePartitionMemory(tempPartition);
  FreePartitionMemoryLight(finerPartition);
  FreePartitionMemoryLight(coarserPartition);

  free(tempPartition);
  free(finerPartition);
  free(coarserPartition);
}

