/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_routines.c - common routines
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/


#include "morsy_routines.h"

#include "stdlib.h"
#include "stdio.h"
#include "memory.h"

int CheckPermutation(int* iperm, int n)
{
  int i;
  int isCorrect = 1;
  int* NumCount = (int*)malloc(sizeof(int)*n);

  memset(NumCount, 0, sizeof(int)*n);

  for(i = 0; i < n; i++)
  {
    NumCount[iperm[i]]++;
  }
  for(i = 0; i < n; i++)
  {
    if (NumCount[i] != 1)
    {
      isCorrect = 0;
    }
  }

  free(NumCount);
  return isCorrect;
}


void GetPermFromIperm(int* iperm, int* perm, int n)
{
  int i;
  for (i = 0; i < n; i++)
    perm[iperm[i]] = i;
}

int LinearSearch(int* Vector, int key, int n)
{
  int index = -1;
  int i;

  for (i = 0; (i < n) && (Vector[i] != key); i++){};
  index = i;

  return index;
}

int GetIntArrayMax(int* Vector, int n)
{
  int i;
  int max;

  max = Vector[0];

  for (i = 1; i < n; i++)
  {
    if (Vector[i] > max)
    {
      max = Vector[i];
    }
  }
  return max;
}

int GetIntArrayMin(int* Vector, int n)
{
  int i;
  int min;

  min = Vector[0];

  for (i = 1; i < n; i++)
  {
    if (Vector[i] > min)
    {
      min = Vector[i];
    }
  }
  return min;
}

int GetDoubleArrayMinIndex(FLOAT_TYPE* Vector, int n)
{
  int i;
  int index = 0;
  FLOAT_TYPE min = Vector[0];

  for(i = 1; i < n; i++)
  {
    if (Vector[i] < min)
    {
      min = Vector[i];
      index = i;
    }
  }
  return index;
}

void GetIntArrayMaxAndMin(int* Vector, int n, int* max, int* min)
{
  int i;

  *max = Vector[0];
  *min = Vector[0];

  for (i = 1; i < n; i++)
  {
    if (Vector[i] < *min)
    {
      *min = Vector[i];
    }
    if (Vector[i] > *max)
    {
      *max = Vector[i];
    }
  }
}



//------------------------------------------------------------------
// BASIC GRAPHS ROUTINES
//------------------------------------------------------------------

void SetNumbersUncompressed(int* iperm, int* Separators, int n)
{
  int i;
  for(i = 0; i < n; i++)
    iperm[Separators[i]] = n - 1 - i;
}

void SetNumbersCompressed(int* iperm, int* Separators,
                          CompressionStructure* Compressed, int nSep, int n)
{
  int i;
  int shift = 0;
  int vertex;
  int* OldNames;
  OldNames = (int*)malloc(sizeof(int)*nSep);

  for(i = 0; i < n; i++)
    if (Compressed->nodeNames[i] > -1)
      OldNames[Compressed->nodeNames[i]] = i;
  
  for(i = 0; i < nSep; i++)
  {
    vertex = OldNames[Separators[i]];
    iperm[vertex] = n - 1 - i - shift;
    shift += Compressed->nCompressedWithIt[vertex];
  }

  for (i = 0; i < Compressed->nCompressed; i++)
  {
    vertex = Compressed->node[i];
    iperm[Compressed->merged[i]] = iperm[vertex] - 
                      Compressed->nCompressedWithIt[vertex];
    Compressed->nCompressedWithIt[vertex]--;
  }
  
  free(OldNames);
}

//-----------------------------------------------------------------------
//BASIC NESTED DISSECTION ROUTINES
//-----------------------------------------------------------------------

int RootedLevelStructure(int root, int* GraphXadj, int* GraphAdjncy, 
                         int* VertexMarks, int* LevelStructure, 
                         int* LevelIndex)
{
  int node;
  int xadjStart;
  int xadjEnd;
  int next;
  int lvlBegin;
  int lvlEnd;
  int currPos;
  int i;
  int j;
  int lvlSize;
  int nLevels;

  nLevels = 0;
  LevelStructure[0] = root; 
  VertexMarks[root] = 0;
  lvlEnd = -1; 
  currPos = 0;
  do
  {
    lvlBegin = lvlEnd + 1;
    lvlEnd = currPos;
    LevelIndex[nLevels] = lvlBegin;
    for (i = lvlBegin; i < lvlEnd + 1; i++)
    {
      xadjStart = GraphXadj[LevelStructure[i]];
      xadjEnd   = GraphXadj[LevelStructure[i] + 1];
      for (j = xadjStart; j < xadjEnd; j++)
      {
        next = GraphAdjncy[j];
        if (VertexMarks[next])
        {
          VertexMarks[next] = 0;
          currPos++;
          LevelStructure[currPos] = next;
        }
      }
    }
    lvlSize = currPos - lvlEnd;
    nLevels++;
  }
  while (lvlSize > 0);

  LevelIndex[nLevels] = lvlEnd + 1;

  for (i = 0; i < lvlEnd + 1; i++) 
    VertexMarks[LevelStructure[i]] = 1;

  return nLevels;
}


int FindPseudoperipheralVertex(int root, int* GraphXadj, int* GraphAdjncy, 
                                int* VertexMarks, int* LevelStructure, 
                                int* LevelIndex)
{
  int i;
  int j;
  int pos;
  int currSize;
  int leaveNumLevels;
  int minDeg;
  int deg;
  int nLevels;

  nLevels = RootedLevelStructure(root, GraphXadj, GraphAdjncy, VertexMarks, 
    LevelStructure, LevelIndex);

  currSize = LevelIndex[nLevels];
  if ((nLevels == 1) || (nLevels == currSize))
    return nLevels;

  do
  {
    pos = LevelIndex[nLevels - 1];
    minDeg = currSize;
    root = LevelStructure[pos];
    if (currSize - 1 != pos)
    {
      for (i = pos; i < LevelIndex[nLevels]; i++)
      {
        deg = 0;
        for (j = GraphXadj[LevelStructure[i]]; j < 
          GraphXadj[LevelStructure[i] + 1]; j++)
        {
          if (VertexMarks[GraphAdjncy[j]]) 
            deg++;
        }
        if (deg < minDeg)
        {
          root = LevelStructure[i];
          minDeg = deg;
        }
      }
    }//if
   leaveNumLevels = RootedLevelStructure(root, GraphXadj, GraphAdjncy, VertexMarks, 
      LevelStructure, LevelIndex);
    if (leaveNumLevels > nLevels)
      nLevels = leaveNumLevels;
    else return nLevels;
  }
  while (nLevels < currSize);
  return nLevels;
}

int FindPseudoperipheralVertexWeighted(int* GraphXadj, int* GraphAdjncy, 
                                int* LevelStructure, int* LevelIndex, int nVertexes)
{
  int i;
  int j;
  int pos;
  int currSize;
  int leaveNumLevels;
  int minDeg;
  int deg;
  int nLevels;
  int root;
  int* VertexMarks;
  
  VertexMarks = (int*)malloc(sizeof(int)*nVertexes);
  for (i = 0; i < nVertexes; i++)
	  VertexMarks[i] = 1;


  nLevels = RootedLevelStructure(0, GraphXadj, GraphAdjncy, VertexMarks, 
    LevelStructure, LevelIndex);

  currSize = LevelIndex[nLevels];
  if ((nLevels == 1) || (nLevels == currSize))
  {
    free(VertexMarks);
    return nLevels;
  }

  do
  {
    pos = LevelIndex[nLevels - 1];
    minDeg = currSize;
    root = LevelStructure[pos];
    if (currSize - 1 != pos)
    {
      for (i = pos; i < LevelIndex[nLevels]; i++)
      {
        deg = GraphXadj[LevelStructure[i] + 1] - GraphXadj[LevelStructure[i]];
        if (deg < minDeg)
        {
          root = LevelStructure[i];
          minDeg = deg;
        }
      }
    }
   leaveNumLevels = RootedLevelStructure(root, GraphXadj, GraphAdjncy, VertexMarks,
     LevelStructure, LevelIndex);

    if (leaveNumLevels > nLevels)
      nLevels = leaveNumLevels;
    else break;
  }
  while (nLevels < currSize);

  free(VertexMarks);

  return nLevels;
}


int GetPartitionLevelSimple(int nLevels, int* LevelIndex)
{
  return nLevels/2;
}


int GetPartitionLevelWeighted(int nLevels, int* LevelStructure, 
                              int* LevelIndex, int* VertexWeights, 
                              int sumWeight)
{
  int i;
  int j;
  int midLevel = nLevels / 2;
  int firstPartWeight = 0;

  for (i = 0; i < nLevels - 2; i++)
  {
    for (j = LevelIndex[i]; j < LevelIndex[i + 1]; j++)
    {
      firstPartWeight += VertexWeights[LevelStructure[j]];
    }
    if (firstPartWeight > sumWeight / 2)
    {
      midLevel = i;
      break;
    }
  }
  return midLevel;
}

int FindSeparator(int vertex, int* GraphAdjncy, int* GraphXadj, 
                  int* VertexMarks, int* LevelStructure, int* LevelIndex,
                  int** Separator, int sepPos, int* nSep, int* nLevels)
{
  int midLevel;
  int nextLevel;
  int node;
  int isSeparator;
  int jstop;
  int i;
  int j;

  *nLevels = FindPseudoperipheralVertex(vertex, GraphXadj, GraphAdjncy, 
    VertexMarks, LevelStructure, LevelIndex);

  (*nSep) = 0;
  if ((*nLevels) <= 2)
  {
    for (i = LevelIndex[0]; i < LevelIndex[*nLevels]; i++)
    {
      (*Separator)[sepPos + (*nSep)] = LevelStructure[i];
      (*nSep)++;
    }
    return 0;
  }

  midLevel = GetPartitionLevelSimple(*nLevels, LevelIndex);
  nextLevel = midLevel + 1;
  for (i = LevelIndex[nextLevel]; i < LevelIndex[nextLevel + 1]; i++)
    GraphXadj[LevelStructure[i]] = - GraphXadj[LevelStructure[i]] - 1;
  
  for (i = LevelIndex[midLevel]; i < LevelIndex[midLevel + 1]; i++)
  {
    node = LevelStructure[i]; 
    isSeparator = 0;

    jstop = (GraphXadj[node + 1] > 0) ?   GraphXadj[node + 1] : 
                                        (-GraphXadj[node + 1] - 1);
    for(j = GraphXadj[node]; j < jstop; j++)
      if (GraphXadj[GraphAdjncy[j]] < 0) 
      {
        isSeparator = 1;
        break;
      }
    if (isSeparator)
    {
      (*Separator)[sepPos + (*nSep)] = node;
      (*nSep)++;
    }
  }
  for (i = LevelIndex[nextLevel]; i < LevelIndex[nextLevel + 1]; i++)
    GraphXadj[LevelStructure[i]] = -GraphXadj[LevelStructure[i]] - 1;
  return midLevel;
}

int FindSeparatorWeighted(GraphStructure* Graph, int* LevelStructure, int* LevelIndex,
                          int* Separator, int* nSep)
{
  int* GraphAdjncy = Graph->Adjncy;
  int* GraphXadj = Graph->Xadj; 
  int* nextLevelVertexes;
  int midLevel;
  int nextLevel;
  int node;
  int isSeparator;
  int jstop;
  int i;
  int j;
  int nVertexes;
  int nLevels;

  nextLevelVertexes = (int*)malloc(sizeof(int)*Graph->nVertexes);
  memset(nextLevelVertexes, 0, sizeof(int)*Graph->nVertexes);

  nLevels = FindPseudoperipheralVertexWeighted(GraphXadj, GraphAdjncy,
    LevelStructure, LevelIndex, Graph->nVertexes);
  
  nVertexes = LevelIndex[nLevels];
  (*nSep) = 0;
  if (nLevels <= 2)
  {
    for (i = LevelIndex[0]; i < Graph->nVertexes; i++)
    {
      Separator[*nSep] = LevelStructure[i];
      (*nSep)++;
    }
    return 0;
  }
 
  midLevel = GetPartitionLevelWeighted(nLevels, LevelStructure, LevelIndex,
    Graph->VertexWeights, Graph->sumWeight);
  nextLevel = midLevel + 1;

  for (i = LevelIndex[nextLevel]; i < LevelIndex[nextLevel + 1]; i++)
  {
    nextLevelVertexes[LevelStructure[i]] = 1;
  }  
  for (i = LevelIndex[midLevel]; i < LevelIndex[midLevel + 1]; i++)
  {
    node = LevelStructure[i]; 
    isSeparator = 0;

    for(j = GraphXadj[node]; j < GraphXadj[node + 1]; j++)
      if (nextLevelVertexes[GraphAdjncy[j]]) 
      {
        isSeparator = 1;
        break;
      }
    if (isSeparator)
    {
      Separator[*nSep] = node;
      (*nSep)++;
    }
  }

  free(nextLevelVertexes);

  return midLevel;
}

int GetNewSubgraphs(int** Subgraphs, int* LevelStructure, int* LevelIndex, int level, 
                    int nSubgr, int nLevels)
{
  int nSubGraphs = 0;
  
  if (nLevels <= 2)
    return nSubGraphs;

  (*Subgraphs)[nSubgr + 1] = LevelStructure[LevelIndex[level - 1]];
  (*Subgraphs)[nSubgr + 2] = LevelStructure[LevelIndex[level + 1]];
  nSubGraphs += 2;

  return nSubGraphs;
}

int PartitionFromLevelStructure(int* LevelStructure, int* LevelIndex, 
                                int nLevels, PartitionStructure* Partition,
                                GraphStructure* Graph)
{
  int nSep = 0;
  int i;
  int j;
  int sepLevelStart = 0;
  int sepLevelEnd;
  int barrier = Graph->sumWeight / 2;
  int weight = 0;
  int vertex;
  int xadjStart;
  int xadjEnd;
  int neighbor;
  int pos;

  for(i = 0; i < Graph->nVertexes; i++)
  {
    Partition->Partition[i] = pSecondPart;
  }

  for (i = 0; (i < nLevels) && (weight < barrier); i++)
  {
    j = LevelIndex[i];
    while ((j < LevelIndex[i + 1]) && (weight < barrier))
    {
      vertex = LevelStructure[j];
      Partition->Partition[vertex] = pFirstPart;
      weight += Graph->VertexWeights[vertex];
      j++;
    }
  }

  return 0;
}

int CheckSeparator(PartitionStructure* Partition,
                                      GraphStructure* Graph)
{
  int error = MORSY_OK;
  int* included = (int*)malloc(sizeof(int)*Graph->nVertexes);
  int i;

  memset(included, 0, sizeof(int)*Graph->nVertexes);

  for(i = 0; i < Partition->nSep; i++)
  {
    included[Partition->SeparatorList[i]]++;
  }

  for(i = 0; i < Partition->nSep; i++)
  {
    if (included[Partition->SeparatorList[i]] != 1)
      error = MORSY_PARTITIONING_ERROR;
  }

  free(included);
  return error;
}

void GetEdgeCut(PartitionStructure* Partition,
                GraphStructure* Graph)
{
  int i;
  int j;
  int xadjStart;
  int xadjEnd;
  
  Partition->edgeCut = 0;
  for(i = 0; i < Graph->nVertexes; i++)
  {
    xadjStart = Graph->Xadj[i];
    xadjEnd   = Graph->Xadj[i + 1];
    for(j = xadjStart; j < xadjEnd; j++)
    {
      if (Partition->Partition[i] != Partition->Partition[Graph->Adjncy[j]])
      {
        Partition->edgeCut++;
      }
    }
  }
  Partition->edgeCut /= 2;
}

int VertexSeparatorFromEdgeSeparator(PartitionStructure* Partition,
                                      GraphStructure* Graph, int isAdjToPartWeightsUsed)
{
  int n = Graph->nVertexes;
  int i;
  int j;
  int neighbor;
  int xadjStart;
  int xadjEnd;
  int pos;
  int part;
  int AdjSum[2];
  int nCutEdges = 0;
  int error = MORSY_OK;

  if (isAdjToPartWeightsUsed)
  {
    pos = 0;
    for (i = 0; i < n; i++)
    {
      part = Partition->Partition[i] - 1;
      if (Partition->AdjToPartWeights[!part][i] > 0)
      {
        Partition->SeparatorList[pos] = i;
        pos++;
      }
    }
  }
  else
  {
    pos = 0;
    for (i = 0; i < n; i++)
    {
      part = Partition->Partition[i] - 1;
      AdjSum[0] = 0;
      AdjSum[1] = 0;
      xadjStart = Graph->Xadj[i];
      xadjEnd  = Graph->Xadj[i + 1];
      for (j = xadjStart; j < xadjEnd; j++)
      {
        neighbor = Graph->Adjncy[j];
        AdjSum[Partition->Partition[neighbor] - 1] += Graph->EdgeWeights[j];
      }
      if (AdjSum[!part] > 0)
      {
        Partition->SeparatorList[pos] = i;
        pos++;
      }
    }
  }
  Partition->nSep = pos;

  for (i = 0; i < Partition->nSep; i++)
  {
    neighbor = Partition->SeparatorList[i];
    Partition->Partition[neighbor] = pSeparator;
  }

  //error = CheckSeparator(Partition, Graph);

  return error;
}

void InitializePartitionParameters(PartitionStructure* Partition, GraphStructure* Graph)
{
  int n = Partition->nVert;
  int part;
  int i;
  int weights[3] = {0, 0, 0};
  int nInPart[3] = {0, 0, 0};

  for (i = 0; i < n; i++)
  {
    part = Partition->Partition[i];
    weights[part] += Graph->VertexWeights[i];
    nInPart[part]++;
  }
  Partition->nFirstPart  = nInPart[pFirstPart];
  Partition->nSecondPart = nInPart[pSecondPart];
  Partition->nSep        = nInPart[pSeparator];
  Partition->firstPartWeight = weights[pFirstPart];
  Partition->secondPartWeight = weights[pSecondPart];
  Partition->sepWeight = weights[pSeparator];
  Partition->sumWeight = Graph->sumWeight;

  if ((Partition->nFirstPart == 0) || (Partition->nSecondPart == 0))
  {
    Partition->disbalance = Graph->sumWeight;
  }
  else
  {
    Partition->disbalance = 
      max((FLOAT_TYPE)Partition->firstPartWeight / (FLOAT_TYPE)Partition->secondPartWeight,
          (FLOAT_TYPE)Partition->secondPartWeight / (FLOAT_TYPE)Partition->firstPartWeight);
  }
}

void UpdateArrayForVertexesList(int* oldPartition, int* newPartition, int* movedVertexes, int nMoved)
{
  int i;
  int vertex;

  for (i = 0; i < nMoved; i++)
  {
    vertex = movedVertexes[i];
    oldPartition[vertex] = newPartition[vertex];
  }
}

