/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_subgraphsCreation.h - finding subgraphs after separator deleting 
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/


#include "morsy_subgraphsCreation.h"

void CalculateAdjncySize(int* Xadj, int* componentsHeads, 
                         int nComponents, int* AdjncySize,
                         DisjointSetList* graphComponents)
{
  int i;
  int vertex;

  for(i = 0; i < nComponents; i++)
  {
    AdjncySize[i] = 0;
    for (vertex = componentsHeads[i]; vertex != -1; 
         vertex = graphComponents->next[vertex])
    {
      AdjncySize[i] += Xadj[vertex + 1] - Xadj[vertex];
    }
  }
}

void CreateSubgraphsStructure(GraphStructure* graph, int* componentsHeads, 
                              int* componentsMap, int nComponents, 
                              DisjointSetList* graphComponents, 
                              GraphStructure** subgraphs, int* partition)
{
  int i;
  int j;
  int k;
  int vertex;
  int pos = 0;
  int maxDegree;
  int maxAdjWeight;
  int xadjStart;
  int xadjEnd;

  for(i = 0; i < nComponents; i++)
  {
    //GlobalMap
    subgraphs[i]->sumWeight = 0;
    subgraphs[i]->maxVertexWeight = 0;
    subgraphs[i]->part = partition[componentsHeads[i]] - 1;
    for (vertex = componentsHeads[i]; vertex != -1; 
         vertex = graphComponents->next[vertex])
    {
      j = componentsMap[vertex];
      subgraphs[i]->GlobalMap[j] = graph->GlobalMap[vertex];
      subgraphs[i]->VertexWeights[j] = graph->VertexWeights[vertex];
      subgraphs[i]->ContractedWeights[j] = graph->ContractedWeights[vertex];
      subgraphs[i]->AdjncyWeights[j] = 0;
      subgraphs[i]->sumWeight += subgraphs[i]->VertexWeights[j];
      if (subgraphs[i]->VertexWeights[j] > subgraphs[i]->maxVertexWeight)
      {
        subgraphs[i]->maxVertexWeight = subgraphs[i]->VertexWeights[j];
      }
    }

    //Xadj, Adjncy
    subgraphs[i]->Xadj[0] = 0;
    pos = 0;
    maxAdjWeight = 0;
    maxDegree = 0;
    for (k = 0, vertex = componentsHeads[i]; vertex != -1; 
      vertex = graphComponents->next[vertex], k++)
    {
      xadjStart = graph->Xadj[vertex];
      xadjEnd   = graph->Xadj[vertex + 1];
      for(j = xadjStart; j < xadjEnd; j++)
      {
        if (partition[graph->Adjncy[j]] != pSeparator)
        {
          subgraphs[i]->Adjncy[pos] = componentsMap[graph->Adjncy[j]];
          subgraphs[i]->EdgeWeights[pos] = graph->EdgeWeights[j];
          subgraphs[i]->AdjncyWeights[k] += 
            subgraphs[i]->VertexWeights[subgraphs[i]->Adjncy[pos]];
          pos++;
        }
      }
      subgraphs[i]->Xadj[k + 1] = pos;
      if (subgraphs[i]->Xadj[k + 1] - subgraphs[i]->Xadj[k] > maxDegree)
      {
        maxDegree = subgraphs[i]->Xadj[k + 1] - subgraphs[i]->Xadj[k];
      }
      if (subgraphs[i]->AdjncyWeights[k] > maxAdjWeight)
      {
        maxAdjWeight = subgraphs[i]->AdjncyWeights[k];
      }
    }
    subgraphs[i]->maxDegree = maxDegree;
    subgraphs[i]->nEdges = pos;
    subgraphs[i]->nVertexes = k; 
    subgraphs[i]->maxAdjWeight = maxAdjWeight;
    subgraphs[i]->isFirst = 0;
  } // for nComponents
}

int FindSubgraphs(GraphStructure* graph, PartitionStructure* partition, 
                  GraphStructure*** subgraphs)
{
  int* componentsHeads;
  int* componentsMap;
  int* AdjncySize;
  int nComponents;
  int i;
  DisjointSetList* graphComponents = 
    (DisjointSetList*)malloc(sizeof(DisjointSetList));

  componentsMap = (int*)malloc(sizeof(int)*graph->nVertexes);
  componentsHeads = (int*)malloc(sizeof(int)*(graph->nVertexes - 
    partition->nSep));

  InitializeDisjointSets(graphComponents, graph->nVertexes);
  ConnectedComponentsFromPartition(graphComponents, graph, 
    partition->Partition);

  nComponents = CreateComponentsMap(graphComponents, partition->Partition, 
    componentsMap, graph->nVertexes, componentsHeads);

  AdjncySize = (int*)malloc(sizeof(int)*nComponents);
  CalculateAdjncySize(graph->Xadj, componentsHeads, nComponents, AdjncySize, 
    graphComponents);
  
  (*subgraphs) = (GraphStructure**)malloc(sizeof(GraphStructure*) * nComponents);
  for(i = 0; i < nComponents; i++)
  {
    (*subgraphs)[i] = (GraphStructure*)malloc(sizeof(GraphStructure));
  }
  for(i = 0; i < nComponents; i++)
  {
    MallocGraphStructure((*subgraphs)[i], 
    graphComponents->setSize[componentsHeads[i]], AdjncySize[i]);
  }

  CreateSubgraphsStructure(graph, componentsHeads, componentsMap, nComponents,
    graphComponents, (*subgraphs), partition->Partition);

  FreeDisjointSet(graphComponents);
  free(graphComponents);
  free(componentsMap);
  free(componentsHeads);
  free(AdjncySize);

  return nComponents;
}

void InitVertexesList(DisjointSetList* graphComponents,
                      ConnectedComponentsStructure* CC,
                      int* Xadj)
{
  int i;
  int vertex;
  int pos = 0;
  int prevPos = 0;
  for(i = 0; i < CC->nComponents; i++)
  {
    CC->AdjSizeInComponent[i] = 0;
    for (vertex = CC->heads[i]; vertex != -1; 
         vertex = graphComponents->next[vertex])
    {
      CC->vertexes[pos++] = vertex;
      CC->AdjSizeInComponent[i] += Xadj[vertex + 1] - Xadj[vertex];
    }
    CC->heads[i] = prevPos;
    prevPos = pos;
  }
  CC->heads[i] = prevPos;
}

int FindConnectedComponents(int* Xadj, int* Adjncy, int nVertexes, 
                            ConnectedComponentsStructure* CC)
{
  int i;
  DisjointSetList* graphComponents = 
    (DisjointSetList*)malloc(sizeof(DisjointSetList));

  InitializeDisjointSets(graphComponents, nVertexes);
  MallocConnectedComponentsStructure(CC, nVertexes);

  ConnectedComponents(graphComponents, Xadj, Adjncy, nVertexes);

  CC->nComponents = CreateGraphComponentsMap(graphComponents, CC);

  if (CC->nComponents > 1)
  {
    CC->AdjSizeInComponent = (int*)malloc(sizeof(int) * CC->nComponents);
    CC->vertexes = (int*)malloc(sizeof(int)*nVertexes);

    InitVertexesList(graphComponents, CC, Xadj);
  }
  else
  {
    CC->heads[0] = 0;
    CC->heads[1] = nVertexes;
  }

  FreeDisjointSet(graphComponents);
  free(graphComponents);

  return CC->nComponents;
}
