/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_types.c - used data structures declaration
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/

#include "morsy_types.h"
#include "stdlib.h"
#include "memory.h"

void MallocAuxilaryGraphStructure(GraphStructure* Graph)
{
  int nVert = Graph->nVertexes;
  
  Graph->EdgeWeights = (int*)malloc(sizeof(int)*Graph->nEdges);
  Graph->ContractedWeights = (int*)malloc(sizeof(int)*nVert);
  Graph->VertexWeights = (int*)malloc(sizeof(int)*nVert);
  Graph->AdjncyWeights = (int*)malloc(sizeof(int)*nVert);
  Graph->Match = (int*)malloc(sizeof(int)*nVert);
  Graph->MatchRoots = (int*)malloc(sizeof(int)*nVert);
  Graph->Map = (int*)malloc(sizeof(int)*nVert);
  Graph->GlobalMap = (int*)malloc(sizeof(int)*nVert);
  memset(Graph->AdjncyWeights, 0, sizeof(int)*nVert);
}

void MallocGraphStructure(GraphStructure* Graph, int nVert, int nEdges)
{
  Graph->nVertexes = nVert;
  Graph->nEdges = nEdges;

  Graph->Xadj = (int*)malloc(sizeof(int)*(nVert + 1));
  Graph->Adjncy = (int*)malloc(sizeof(int)*nEdges);
  
  MallocAuxilaryGraphStructure(Graph);
}

void MallocGraphStructureShort(GraphStructure* Graph, int nVert, int nEdges)
{
  Graph->nVertexes = nVert;
  Graph->nEdges = nEdges;
  Graph->Xadj = (int*)malloc(sizeof(int)*(nVert + 1));
  Graph->Adjncy = (int*)malloc(sizeof(int)*nEdges);
  Graph->GlobalMap = (int*)malloc(sizeof(int)*nVert);
}

void InitializeGraphStructure(GraphStructure* Graph, int* Xadj, 
                                         int* Adjncy, int nVert)
{
  int i;
  
  Graph->nVertexes = nVert;
  Graph->nEdges = Xadj[nVert];

  Graph->Xadj = Xadj;
  Graph->Adjncy = Adjncy;
  
  MallocAuxilaryGraphStructure(Graph);

  Graph->maxDegree = 0;
  for(i = 0; i < nVert; i++)
  {
    Graph->GlobalMap[i] = i;
    if (Graph->Xadj[i + 1] - Graph->Xadj[i] > Graph->maxDegree)
    {
      Graph->maxDegree = Graph->Xadj[i + 1] - Graph->Xadj[i];
    }
  }

  Graph->isFirst = 1;
}

void InitializeDisconnectedGraphStructure(GraphStructure** Graph, 
                                          ConnectedComponentsStructure* CC,
                                          int* Xadj, int* Adjncy)
{
  int maxGraphID = 0;
  int i;
  int j;
  int k;
  int pos;
  int vertex;
  int itsNum;
  int neighbor;

  for (i = 0; i < CC->nComponents; i++)
  {
    MallocGraphStructure(Graph[i], CC->heads[i+1] - CC->heads[i], 
      CC->AdjSizeInComponent[i]);

    Graph[i]->Xadj[0] = 0;
    pos = 0;
    for(j = CC->heads[i]; j < CC->heads[i + 1]; j++)
    {
      vertex = CC->vertexes[j];
      CC->vertexComponentID[vertex] = i;
      itsNum = j - CC->heads[i];
      for(k = Xadj[vertex]; k < Xadj[vertex + 1]; k++)
      {
        Graph[i]->Adjncy[pos++] = CC->map[Adjncy[k]];
      }
      Graph[i]->Xadj[itsNum + 1] = pos;
      Graph[i]->GlobalMap[itsNum] = vertex;
      if (Graph[i]->Xadj[itsNum + 1] - Graph[i]->Xadj[itsNum] > Graph[i]->maxDegree)
      {
        Graph[i]->maxDegree = Graph[i]->Xadj[itsNum + 1] - Graph[i]->Xadj[itsNum];
      }
    }
    Graph[i]->isFirst = 0;

    if (Graph[i]->nVertexes > Graph[maxGraphID]->nVertexes)
      maxGraphID = i;
  }
  Graph[maxGraphID]->isFirst = 1;
}

void InitializeCompressedWeights(GraphStructure* Graph, CompressionStructure* Compressed)
{
  int i;
  int j;
  int xadjStart;
  int xadjEnd;
  int name;
  int vertex;
  Graph->maxVertexWeight = 1;
  Graph->maxAdjWeight = 0;
  Graph->sumWeight = Compressed->nOriginal;
  for (i = 0; i < Compressed->nOriginal; i++)
  {
    vertex = i;
    name = Compressed->nodeNames[vertex];
    if (name > -1)
    {
      Graph->VertexWeights[name] = Compressed->nCompressedWithIt[vertex] + 1;
      if (Graph->VertexWeights[name] > Graph->maxVertexWeight)
      {
        Graph->maxVertexWeight = Graph->VertexWeights[name];
      }
    }
  }

  for(i = 0; i < Graph->nVertexes; i++)
  {
    xadjStart = Graph->Xadj[i];
    xadjEnd   = Graph->Xadj[i + 1];
    Graph->AdjncyWeights[i] = 0;
    for(j = xadjStart; j < xadjEnd; j++)
    {
      Graph->AdjncyWeights[i] += Graph->VertexWeights[Graph->Adjncy[j]];
    }
    if (Graph->AdjncyWeights[i] > Graph->maxAdjWeight)
    {
      Graph->maxAdjWeight = Graph->AdjncyWeights[i]; 
    }
  }

  for(i = 0; i < Graph->nEdges; i++)
  {
    Graph->EdgeWeights[i] = 1;
  }
  memset(Graph->ContractedWeights, 0, sizeof(int)*Graph->nVertexes);
}

void InitializeCompressedDisconnectedWeights(GraphStructure** Graphs, 
                                             CompressionStructure* Compressed, 
                                             ConnectedComponentsStructure* graphCC)
{
  int i;
  int j;
  int xadjStart;
  int xadjEnd;
  int name;
  int vertex;
  int componentID;
  GraphStructure* Graph;

  for (i = 0; i < Compressed->nOriginal; i++)
  {
    name = Compressed->nodeNames[i];
    if (name > -1)
    {
      componentID = graphCC->vertexComponentID[name];
      vertex = graphCC->map[name];
      Graphs[componentID]->VertexWeights[vertex] = Compressed->nCompressedWithIt[i] + 1;
    }
  }

  for (componentID = 0; componentID < graphCC->nComponents; componentID++)
  {
    Graph = Graphs[componentID];

    Graph->maxVertexWeight = 0;
    Graph->maxAdjWeight = 0;
    Graph->sumWeight = 0;
    for(i = 0; i < Graph->nVertexes; i++)
    {
      xadjStart = Graph->Xadj[i];
      xadjEnd   = Graph->Xadj[i + 1];
      Graph->AdjncyWeights[i] = 0;
      for(j = xadjStart; j < xadjEnd; j++)
      {
        Graph->AdjncyWeights[i] += Graph->VertexWeights[Graph->Adjncy[j]];
      }
      if (Graph->AdjncyWeights[i] > Graph->maxAdjWeight)
      {
        Graph->maxAdjWeight = Graph->AdjncyWeights[i]; 
      }
      if (Graph->VertexWeights[i] > Graph->maxVertexWeight)
      {
        Graph->maxVertexWeight = Graph->VertexWeights[i];
      }
      Graphs[componentID]->sumWeight += Graphs[componentID]->VertexWeights[i];
    }

    for(i = 0; i < Graph->nEdges; i++)
    {
      Graph->EdgeWeights[i] = 1;
    }
    memset(Graph->ContractedWeights, 0, sizeof(int)*Graph->nVertexes);
  }
}


void InitializeUncompressedWeights(GraphStructure* Graph)
{
  int i;
  for(i = 0; i < Graph->nVertexes; i++)
  {
    Graph->VertexWeights[i] = 1;
    Graph->AdjncyWeights[i] = Graph->Xadj[i + 1] - Graph->Xadj[i];
  }
  Graph->maxVertexWeight = 1;
  Graph->maxAdjWeight = Graph->maxDegree;
  Graph->sumWeight = Graph->nVertexes;

  for(i = 0; i < Graph->nEdges; i++)
  {
    Graph->EdgeWeights[i] = 1;
  }
  memset(Graph->ContractedWeights, 0, sizeof(int)*Graph->nVertexes);
}

void FreeGraphStructure(GraphStructure* Graph)
{
  free(Graph->Adjncy);
  free(Graph->Xadj);
  free(Graph->GlobalMap);
  free(Graph->EdgeWeights);
  free(Graph->VertexWeights);
  free(Graph->AdjncyWeights);
  free(Graph->Map);
  free(Graph->Match);
  free(Graph->MatchRoots);
  free(Graph->ContractedWeights);
}

void CopyGraphStructure(GraphStructure* destination, GraphStructure* source)
{
  destination->Adjncy = source->Adjncy;
  destination->AdjncyWeights = source->AdjncyWeights;
  destination->GlobalMap = source->GlobalMap;
  destination->Map = source->Map;
  destination->Match = source->Match;
  destination->MatchRoots = source->MatchRoots;
  destination->maxAdjWeight = source->maxAdjWeight;
  destination->maxDegree = source->maxDegree;
  destination->maxVertexWeight = source->maxVertexWeight;
  destination->nMatchRoots = source->nMatchRoots;
  destination->nVertexes = source->nVertexes;
  destination->nEdges = source->nEdges;
  destination->part = source->part;
  destination->sumWeight = source->sumWeight;
  destination->VertexWeights = source->VertexWeights;
  destination->EdgeWeights = source->EdgeWeights;
  destination->Xadj = source->Xadj;
  destination->isFirst = source->isFirst;
}

void MallocConnectedComponentsStructure(ConnectedComponentsStructure* CC, 
                                        int nVertexes)
{
  CC->nAllVertexes = nVertexes;
  CC->heads = (int*)malloc(sizeof(int)*nVertexes);
  CC->map = (int*)malloc(sizeof(int)*nVertexes);
  CC->vertexComponentID = (int*)malloc(sizeof(int)*nVertexes);
}

void FreeConnectedComponentsStructure(ConnectedComponentsStructure* CC)
{
  if (CC->nComponents > 1)
  {
    free(CC->AdjSizeInComponent);
    free(CC->vertexes);
  }
  free(CC->heads);
  free(CC->map);
  free(CC->vertexComponentID);
}

