/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* morsy_types.h - used data structures declaration
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*/


#ifndef __PMORSY_TYPES_H__
#define __PMORSY_TYPES_H__

#include "pmorsy.h"

#ifdef MORSY_TIMERS
#include "morsy_timer.h"
#endif


#ifdef MORSY_USE_MKL

#include "mkl_vsl.h"
typedef VSLStreamStatePtr MORSY_RNG_PTR;

#else

typedef void* MORSY_RNG_PTR;

#endif


#if MORSY_PRECISION == DOUBLE
#define EPSILON 1E-12
#define FLOAT_TYPE double
#define pfabs fabs
#else
#define EPSILON 1E-7
#define FLOAT_TYPE float
#define pfabs fabsf
#endif


struct 
{ 
  int* merged;            
  int* node;              
  int* nCompressedWithIt;
  int* nodeNames;
  int nCompressed;        
  int nOriginal;          
} typedef CompressionStructure;

struct 
{
  int* Adjncy;         
  int* Xadj;             
  int* VertexWeights;
  int* EdgeWeights;
  int* ContractedWeights;
  int* AdjncyWeights;  
  int* Match;          
  int* MatchRoots;
  int nMatchRoots;
  int* Map;            
  int* GlobalMap;      
  int nVertexes;       
  int nEdges;          
  int sumWeight;       
  int maxDegree;       
  int maxVertexWeight; 
  int maxAdjWeight;     
  int part;
  int isFirst;
} typedef GraphStructure;

typedef enum { 
  pSeparator, 
  pFirstPart,  
  pSecondPart
} PartitionVertexType;

struct 
{
  int prev;                     
  int next;                     
} typedef DoublyLinkedList;

struct
{
  int* Buckets;                 
  DoublyLinkedList* Nodes;      
  int maxGain;                  
  int nNodes;                   
  int minPossGain;              
  int nBuckets;                 
  int* gainValue;               
  int gainCut;
  int oldMaxGain;
} typedef GainsStructure;

struct
{
  int* Partition; 
  int* SeparatorList;             
  GainsStructure* Gains[2];       
  int nSep;                       
  int nVert;                      
  int nFirstPart;                 
  int nSecondPart;                
  int firstPartWeight;            
  int secondPartWeight;           
  int sepWeight;                  
  int nGains;                     
  int sumWeight;
  int* SeparatorListPos;
  int* AdjToPartWeights[2];       
  FLOAT_TYPE disbalance;
  int edgeCut;
} typedef PartitionStructure;

struct
{
  int* vertexComponentID;
  int* heads;
  int* vertexes;
  int* map;
  int* AdjSizeInComponent;
  int nComponents;
  int nAllVertexes;
} typedef ConnectedComponentsStructure;


typedef int (*MatchingAlgorithm)(GraphStructure*, 
                      int*, int*, MORSY_RNG_PTR);


typedef void (*PartitioningAlgorithm)(PartitionStructure*,  
                                GraphStructure*,  MORSY_RNG_PTR);

struct 
{
  int NDType;
  int matchingType;
  int isCompress;             
  MatchingAlgorithm Matching;
  PartitioningAlgorithm Partitioning;
  int nMinVertexes;           
  int nMaxCoarseningSteps;    
  FLOAT_TYPE disbalanceCoeff; 
  int gainCut;                 
  int refinementStep;
  int nComponents;
  int refineLimit;
  int nThreads;
  FLOAT_TYPE disbalanceStop;
  int gainCutLimit;
  int barrier;
  MORSY_RNG_PTR* randomStream;
  #ifdef MORSY_TIMERS
  MORSy_timer timer[N_TIMERS];
  #endif
} typedef NestedDissectionParameters;

struct TaskStructure
{
  GraphStructure* graph;
  int* Separators;
  int nSep;
  GraphStructure** subgraphs;
  int nSubgraphs;
  int taskID;
  struct TaskStructure** subTasks;
  int part;
  int level;
};

typedef struct TaskStructure TaskStructure;

typedef void (*AshcraftLiuWeighted)(GraphStructure*, PartitionStructure*, 
                                    PartitionStructure*, 
                                    NestedDissectionParameters*, int*, int*, int);

void MallocGraphStructure(GraphStructure* Graph, int nVert, int nEdges);

void MallocGraphStructureShort(GraphStructure* Graph, int nVert, int nEdges);

void InitializeGraphStructure(GraphStructure* Graph, int* Xadj, 
                                         int* Adjncy, int nVert);

void FreeGraphStructure(GraphStructure* Graph);

void CopyGraphStructure(GraphStructure* destination, GraphStructure* source);

void MallocConnectedComponentsStructure(ConnectedComponentsStructure* CC, 
                                        int nVertexes);
void FreeConnectedComponentsStructure(ConnectedComponentsStructure* CC);

void InitializeDisconnectedGraphStructure(GraphStructure** Graph, 
                                          ConnectedComponentsStructure* CC,
                                          int* Xadj, int* Adjncy);

void InitializeCompressedWeights(GraphStructure* Graph, CompressionStructure* Compressed);

void InitializeCompressedDisconnectedWeights(GraphStructure** Graphs, 
                                             CompressionStructure* Compressed, 
                                             ConnectedComponentsStructure* graphCC);

void InitializeUncompressedWeights(GraphStructure* Graph);


#endif