/*
* Copyright (C) 2015 University of Nizhni Novgorod
* This file is part of the PMORSy ordering library, v 1.2.0
* 
* pmorsy.c - to start reordering
* 
* This file may be used under the terms of the GNU Lesser General Public 
* License version 2.1 as published by the Free Software Foundation 
* and appearing in the file LICENSE.LGPL included in the packaging of this 
* file. Please review the following information to ensure the GNU Lesser 
* General Public License version 2.1 requirements will be met: 
* http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
*
*/


#include "pmorsy.h"
#include "morsy_AshcraftLiu.h"
#include "morsy_routines.h"
#include "morsy_types.h"
#include "morsy_coarsening.h"
#include "morsy_compression.h"
#include "morsy_multilevelND.h"
#include "morsy_NDBasic.h"
#include "morsy_subgraphsCreation.h"
#include "morsy_randomGenerator.h"
#include "omp.h"
#include <stdio.h>

void MORSY_SetUpOptions(int* options, NestedDissectionParameters* NDParameters)
{
  if (in(options[MORSY_IS_COMPRESS], 0, 1))
  {
    NDParameters->isCompress = options[MORSY_IS_COMPRESS];
  }
  else
  {
    NDParameters->isCompress = 1;
  }

  if (in(options[MORSY_MATCHING_TYPE], 0, 2))
  {
    switch (options[MORSY_MATCHING_TYPE])
    {
    case 0: NDParameters->Matching = RandomMatching; break;
    case 1: NDParameters->Matching = HeavyEdgeMatching; break;
    case 2: NDParameters->Matching = HeavyCliqueMatching; break;
    }
    NDParameters->matchingType = options[MORSY_MATCHING_TYPE];
  }
  else
  {
    NDParameters->Matching = HeavyEdgeMatching;
    NDParameters->matchingType = 1;
  }

  if (in(options[MORSY_PARTITIONING_TYPE], 0, 3))
  {
    switch (options[MORSY_PARTITIONING_TYPE])
    {
    case 0: NDParameters->Partitioning = BFSPartition; break;
    case 1: NDParameters->Partitioning = GraphGrowingPartition; break;
    case 2: NDParameters->Partitioning = GreedyGraphGrowingPartition; break;
    case 3: NDParameters->Partitioning = KerniganLinPartition; break;
    }
  }
  else
  {
    NDParameters->Partitioning = BFSPartition;
  }
  
  if (options[MORSY_COARSE_TO] > 3)
  {
    NDParameters->nMinVertexes = options[MORSY_COARSE_TO];
  }
  else
  {
    NDParameters->nMinVertexes = 20;
  }

  if (options[MORSY_COARSERNING_STEPS] > -1)
  {
    NDParameters->nMaxCoarseningSteps = options[MORSY_COARSERNING_STEPS];
  }
  else
  {
    NDParameters->nMaxCoarseningSteps = 15;
  }

  if (in(options[MORSY_DISBALANCE_COEFF], 0, 100))
  {
    NDParameters->disbalanceCoeff = (FLOAT_TYPE)options[MORSY_DISBALANCE_COEFF]/100.0;
  }
  else
  {
    NDParameters->disbalanceCoeff = 0.2;
  }

  if (in(options[MORSY_REFINEMENT_STEP], 1, 5))
  {
    NDParameters->refinementStep = options[MORSY_REFINEMENT_STEP];
  }
  else
  {
    NDParameters->refinementStep = 1;
  }
  if (in(options[MORSY_IS_CONNECTED], 0, 2))
  {
    NDParameters->nComponents = options[MORSY_IS_CONNECTED];
  }
  else
  {
    NDParameters->nComponents = 0;
  }
  if (options[MORSY_IS_REFINEMENT_LIMITED] > 0)
  {
    NDParameters->refineLimit = options[MORSY_IS_REFINEMENT_LIMITED];
  }
  else
  {
    NDParameters->refineLimit = 0;
  }
  if (in(options[MORSY_REFINEMENT_STOP], 10, 50))
  {
    NDParameters->disbalanceStop = (FLOAT_TYPE)options[MORSY_REFINEMENT_STOP] / 10.0;
  }
  else
  {
    NDParameters->disbalanceStop = 1.5;
  }
  if (in(options[MORSY_REFINEMENT_CUT], 1, 3))
  {
    NDParameters->gainCutLimit = options[MORSY_REFINEMENT_CUT];
  }
  else
  {
    NDParameters->gainCutLimit = 3;
  }
  if (in(options[MORSY_REFINEMENT_TYPE], 0, 1))
  {
    NDParameters->NDType = options[MORSY_REFINEMENT_TYPE];
  }
  else
  {
    NDParameters->NDType = 1;
  }
  if (in(options[MORSY_BARRIER], 0, 3000))
  {
    NDParameters->barrier = options[MORSY_BARRIER];
  }
  else
  {
    NDParameters->barrier = 1000;
  }
  if (options[MORSY_NUM_THREADS] > 0)
  {
    NDParameters->nThreads = options[MORSY_NUM_THREADS];
  }
  else
  {
    NDParameters->nThreads = 1;
  }

}

void MORSY_SetDefaultOptions(int* options)
{
  options[MORSY_IS_COMPRESS] = 1;
  options[MORSY_MATCHING_TYPE] = 1;
  options[MORSY_COARSE_TO] = 100;
  options[MORSY_COARSERNING_STEPS] = 10;
  options[MORSY_DISBALANCE_COEFF] = 20;
  options[MORSY_REFINEMENT_STEP] = 1;
  options[MORSY_IS_CONNECTED] = 0;
  options[MORSY_REFINEMENT_STOP] = 20;
  options[MORSY_REFINEMENT_CUT] = 3;
  options[MORSY_NUM_THREADS] = 4;
  options[MORSY_REFINEMENT_TYPE] = 1;
  options[MORSY_BARRIER] = 1000;
  options[MORSY_PARTITIONING_TYPE] = 0;
  options[MORSY_NUM_THREADS] = 1;
}

void MORSY_SetTimePriorityOptions(int* options)
{
  MORSY_SetDefaultOptions(options);
}

void MORSY_SetQualityPriorityOptions(int* options)
{
  options[MORSY_IS_COMPRESS] = 1;
  options[MORSY_MATCHING_TYPE] = 1;
  options[MORSY_COARSE_TO] = 20;
  options[MORSY_COARSERNING_STEPS] = 10;
  options[MORSY_DISBALANCE_COEFF] = 20;
  options[MORSY_REFINEMENT_STEP] = 1;
  options[MORSY_IS_CONNECTED] = 0;
  options[MORSY_NUM_THREADS] = 4;
  options[MORSY_REFINEMENT_TYPE] = 0;
  options[MORSY_BARRIER] = 1000;
  options[MORSY_PARTITIONING_TYPE] = 0;
  options[MORSY_NUM_THREADS] = 1;
}

void MORSY_PrintOptions(int* options)
{
  printf("*******************************\n");
  printf("MORSy reordering options:\n");
  printf("is compress: %d\n", options[MORSY_IS_COMPRESS]);
  
  switch (options[MORSY_MATCHING_TYPE])
  {
  case 0: printf("matching: RandomMatching\n"); break;
  case 1: printf("matching: HeavyEdgeMatching\n"); break;
  case 2: printf("matching: HeavyCliqueMatching\n"); break;
  }

  switch(options[MORSY_PARTITIONING_TYPE])
  {
  case 0: printf("init partitioning: BFS Partition\n"); break;
  case 1: printf("init partitioning: GraphGrowingPartition\n"); break;
  case 2: printf("init partitioning: GreedyGraphGrowingPartition\n"); break;
  case 3: printf("init partitioning: KerniganLinPartition\n"); break;
  }

  printf("number of coarsening steps: %d\n", options[MORSY_COARSERNING_STEPS]);
  printf("coarsen to: %d\n", options[MORSY_COARSE_TO]);
  printf("refinement step: %d\n", options[MORSY_REFINEMENT_STEP]);
  printf("disbalance coefficient: %.2f\n", (FLOAT_TYPE)options[MORSY_DISBALANCE_COEFF] / 100.0);
  printf("is refinement limited: %d\n", options[MORSY_IS_REFINEMENT_LIMITED]);
  printf("refinement disbalance stop: %.2f\n", (FLOAT_TYPE)options[MORSY_REFINEMENT_STOP] / 10.0);
  printf("refinement gainCut limit: %d\n", options[MORSY_REFINEMENT_CUT]);
  printf("omp task barrier: %d\n", options[MORSY_BARRIER]);
  printf("*******************************\n");
}


void MORSY_PrintOptions_Inner(NestedDissectionParameters* NDParameters, int* options)
{
  printf("*******************************\n");
  printf("MORSy reordering options:\n");
  printf("is compress: %d\n", NDParameters->isCompress);
  printf("matching: %d\n", options[MORSY_MATCHING_TYPE]);
  printf("init partitioning: %d\n", options[MORSY_PARTITIONING_TYPE]);
  printf("number of coarsening steps: %d\n", NDParameters->nMaxCoarseningSteps);
  printf("coarsen to: %d\n", NDParameters->nMinVertexes);
  printf("refinement type: %d\n", NDParameters->NDType);
  printf("refinement step: %d\n", NDParameters->refinementStep);
  printf("disbalance coefficient: %.2f\n", NDParameters->disbalanceCoeff);
  printf("is refinement limited: %d\n", NDParameters->refineLimit);
  printf("refinement disbalance stop: %.2f\n", NDParameters->disbalanceStop);
  printf("refinement gainCut limit: %d\n", NDParameters->gainCutLimit);
  printf("*******************************\n");
}


#ifdef MORSY_TIMERS
void MORSy_PrintTime_Inner(MORSy_timer* timer)
{
  printf("*******************************\n");
  printf("PMORSy time (seconds):\n");
  printf("Compression: %lf\n", timer[MORSY_TIME_COMPRESSION].time);
  printf("Coarsening: %lf\n", timer[MORSY_TIME_COARSENING].time);
  printf("- Matching: %lf\n", timer[MORSY_TIME_MATCHING].time);
  printf("- Contracting: %lf\n", timer[MORSY_TIME_CONTRACTING].time);
  printf("Initial partitioning: %lf\n", timer[MORSY_TIME_INIT_PARTITIONING].time);
  printf("Uncoarsening: %lf\n", timer[MORSY_TIME_UNCOARSENING].time);
  printf("- Refinement: %lf\n", timer[MORSY_TIME_REFINEMENT].time);
  printf("- Projection: %lf\n", timer[MORSY_TIME_PROJECTION].time);
  printf("Splitting: %lf\n", timer[MORSY_TIME_SPLITTING].time);
  printf("Basic ND: %lf\n", timer[MORSY_TIME_LEAFS].time);
  printf("First task: %lf\n", timer[MORSY_TIME_FIRST_TASK].time);
  printf("- Coarsening: %lf\n", timer[MORSY_TIME_FIRST_TASK_COARSENING].time);
  printf("-- Matching: %lf\n", timer[MORSY_TIME_FIRST_TASK_MATCHING].time);
  printf("- Initial partitioning: %lf\n", timer[MORSY_TIME_FIRST_TASK_INIT_PARTITIONING].time);
  printf("- Uncoarsening: %lf\n", timer[MORSY_TIME_FIRST_TASK_UNCOARSENING].time);
  printf("-- Refinement: %lf\n", timer[MORSY_TIME_FIRST_TASK_REFINEMENT].time);
  printf("Total time: %lf\n", timer[MORSY_TIME_SUM].time);
  printf("*******************************\n");
}
#endif


int PMORSY_NestedDissection(int n, int* xadj, int* adjncy, int* perm, int* iperm, int* options)
{
  NestedDissectionParameters NDParameters;
  AshcraftLiuWeighted AshcraftLiuProcedure;
  int* GraphXadj;
  int* GraphAdjncy;
  int* Separators;
  int nCompressed = 0;
  int nVert = n;
  int i;
  int j;
  int isConnected;
  CompressionStructure* Compressed;  
  GraphStructure** Graph;
  ConnectedComponentsStructure* graphCC;

#ifdef MORSY_TIMERS
  InitTimers(NDParameters.timer);
	StartTimer(&NDParameters.timer[MORSY_TIME_SUM]);
#endif
  Compressed = 
    (CompressionStructure*)malloc(sizeof(CompressionStructure));
  graphCC = 
    (ConnectedComponentsStructure*)malloc(sizeof(ConnectedComponentsStructure));

  Separators = (int*)malloc(sizeof(int)*n);

  if (options == NULL)
  {
    options = (int*)malloc(sizeof(int)*MORSY_NUM_OPTIONS);
    MORSY_SetDefaultOptions(options);
  }
  else if (options[MORSY_IS_DEFAULT] == 1)
  {
    MORSY_SetDefaultOptions(options);
  }
  MORSY_SetUpOptions(options, &NDParameters);
  switch(options[MORSY_REFINEMENT_TYPE])
  {
  case 1: AshcraftLiuProcedure = AshcraftLiuProcedureOuterWeighted; break;
  case 0: AshcraftLiuProcedure = AshcraftLiuProcedureOuterWeightedUnlim; break;
  }
  omp_set_num_threads(NDParameters.nThreads);

#ifdef MORSY_PRINT
    MORSY_PrintOptions_Inner(&NDParameters, options);
#endif
  
  InitializeRandomGenerators(&NDParameters.randomStream, NDParameters.nThreads);

  if (NDParameters.isCompress)
  {
#ifdef MORSY_TIMERS
    StartTimer(&(NDParameters.timer[MORSY_TIME_COMPRESSION]));
#endif
    nCompressed = GraphCompressionProcedure(adjncy, xadj, 
      &GraphAdjncy, &GraphXadj, Compressed, n);
    nVert = n - nCompressed;
#ifdef MORSY_TIMERS
    FinishTimer(&(NDParameters.timer[MORSY_TIME_COMPRESSION]));
#endif
  }
  else
  {
    GraphXadj = (int*)malloc(sizeof(int) * (n + 1));
    memcpy(GraphXadj, xadj,  sizeof(int) * (n + 1));
    GraphAdjncy = (int*)malloc( sizeof(int) * xadj[n]);
    memcpy(GraphAdjncy, adjncy, sizeof(int) * xadj[n]);
  }

  if (NDParameters.nComponents == 0)
  {
    NDParameters.nComponents = FindConnectedComponents(GraphXadj, GraphAdjncy, 
      nVert, graphCC);
    isConnected = 0;
  }
  else
  {
    NDParameters.nComponents = 1;
    graphCC->nComponents = 1;
    graphCC->heads = (int*)malloc(sizeof(int)*2);
    graphCC->heads[0] = 0;
    graphCC->heads[1] = nVert;
    isConnected = 1;
  }

  Graph = (GraphStructure**)malloc(sizeof(GraphStructure*) * graphCC->nComponents);
  for(i = 0; i < graphCC->nComponents; i++)
  {
    Graph[i] = (GraphStructure*)malloc(sizeof(GraphStructure));
  }


#ifdef MORSY_PRINT
    printf("nCompressed: %d\n", nCompressed);
    printf("new number of vertexes: %d\n", n - nCompressed);
    printf("new number of edges: %d\n", GraphXadj[n - nCompressed]);
    printf("number of graph components: %d\n", graphCC->nComponents);
#endif

  if (NDParameters.nComponents == 1)
  {
    InitializeGraphStructure(Graph[0], GraphXadj, GraphAdjncy, nVert);
    if (nCompressed)
    {
      InitializeCompressedWeights(Graph[0], Compressed);
    }
    else
    {
      InitializeUncompressedWeights(Graph[0]);
    }
  }
  else
  {
    InitializeDisconnectedGraphStructure(Graph, graphCC, GraphXadj, GraphAdjncy);
    if (nCompressed)
    {
      InitializeCompressedDisconnectedWeights(Graph, Compressed, graphCC);
    }
    else
    {
      for (i = 0; i < graphCC->nComponents; i++)
      {
        InitializeUncompressedWeights(Graph[i]);
      }
    }
  }

  for (i = 0; i < graphCC->nComponents; i++)
  {
    MultilevelNestedDissection(Graph[i], Separators + graphCC->heads[i], 
      AshcraftLiuProcedure, &NDParameters);
  }

  if (!nCompressed)
  {
    SetNumbersUncompressed(iperm, Separators, n);
  }
  else
  {
    SetNumbersCompressed(iperm, Separators, Compressed, nVert, n);
  }

  if (NDParameters.isCompress)
  {
    FreeCompressionStructure(Compressed);
  }

  FinalizeRandomGenerators(&NDParameters.randomStream, NDParameters.nThreads);

  if (perm != NULL)
    GetPermFromIperm(iperm, perm, n);

  free(Separators);
  free(Compressed);
  if (isConnected == 0)
  {
    FreeConnectedComponentsStructure(graphCC);
  }
  else
  {
    free(graphCC->heads);
  }
  free(graphCC);

  free(Graph);

#ifdef MORSY_TIMERS
  FinishTimer(&NDParameters.timer[MORSY_TIME_SUM]);
  FinilizeTimers(NDParameters.timer);
#endif

#ifdef MORSY_TIMERS
#ifdef MORSY_PRINT
  MORSy_PrintTime_Inner(NDParameters.timer);
#endif
#endif
  if (CheckPermutation(iperm, n))
    return MORSY_OK;
  else 
    return MORSY_ERROR;
}
